# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* Binary Execution Wrappers **************************************************************

import concurrent.futures
import copy
import ctypes
import numpy as np
from inspect import isgenerator
import os
import sys
import time
from typing import Callable, Optional
import weakref


from psyneulink.core import llvm as pnlvm
from psyneulink.core.globals.context import Context

from . import builder_context, jit_engine, scheduler
from .debug import debug_env

__all__ = ['CompExecution', 'FuncExecution', 'MechExecution']


def _tupleize(x):
    try:
        return tuple(_tupleize(y) for y in x)
    except TypeError:
        return x if x is not None else tuple()


def _pretty_size(size):
    units = ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB']
    for u in units:
        if abs(size) > 1536.0:
            size /= 1024.0
        else:
            break

    return "{:.2f} {}".format(size, u)


class Execution:
    def __init__(self):
        self._debug_env = debug_env

    def _get_compilation_param(self, name, init_method, arg):
        saved = getattr(self, name, None)
        if saved is None:
            struct_ty = self._bin_func.byref_arg_types[arg]
            init_f = getattr(self._obj, init_method)
            init_start = time.time()
            initializer = init_f(self._execution_context)

            init_end = time.time()
            struct = struct_ty(*initializer)
            struct_end = time.time()

            # numpy "frombuffer" creates a shared memory view of the provided buffer
            numpy_struct = np.frombuffer(struct, dtype=self._bin_func.np_arg_dtypes[arg], count=1)

            assert numpy_struct.nbytes == ctypes.sizeof(struct), \
                "Size mismatch ({}), numpy: {} vs. ctypes:{}".format(name, numpy_struct.nbytes, ctypes.sizeof(struct))

            saved = numpy_struct
            setattr(self, name, saved)

            if "time_stat" in self._debug_env:
                print("Time to get initializer for struct:", name,
                      "for", self._obj.name, ":", init_end - init_start)
                print("Time to instantiate struct:", name,
                      "for", self._obj.name, ":", struct_end - init_end)

            if "stat" in self._debug_env:
                print("Instantiated struct:", name, "( size:" ,
                      _pretty_size(ctypes.sizeof(struct_ty)), ")",
                      "for", self._obj.name)

            numpy_struct.shape = ()

            if name == '_state':
                self._copy_params_to_pnl(self._execution_context,
                                         self._obj,
                                         numpy_struct,
                                         "llvm_state_ids")

            elif name == '_param':
                self._copy_params_to_pnl(self._execution_context,
                                         self._obj,
                                         numpy_struct,
                                         "llvm_param_ids")

        return saved

    def _copy_params_to_pnl(self, context, component, params, ids:str):

        assert len(params.dtype.names) == len(getattr(component, ids))

        for numpy_name, attribute in zip(params.dtype.names, getattr(component, ids)):

            numpy_field = params[numpy_name]
            assert numpy_field.base is params or numpy_field.base is params.base

            def _enumerate_recurse(elements):
                for numpy_element_name, element in zip(numpy_field.dtype.names, elements):
                    numpy_element = numpy_field[numpy_element_name]
                    assert numpy_element.base is numpy_field.base

                    self._copy_params_to_pnl(context=context,
                                             component=element,
                                             params=numpy_element,
                                             ids=ids)

            # Handle custom compiled-only structures by name
            if attribute == 'nodes':
                _enumerate_recurse(component._all_nodes)

            elif attribute == 'projections':
                _enumerate_recurse(component._inner_projections)

            elif attribute == '_parameter_ports':
                _enumerate_recurse(component._parameter_ports)

            else:
                # TODO: Reconstruct Python RandomState
                if attribute == "random_state":
                    continue

                # TODO: Reconstruct Python memory storage
                if attribute == "ring_memory":
                    continue

                # TODO: Reconstruct Time class
                if attribute == "num_executions":
                    continue

                # TODO: Add support for syncing optimizer state
                if attribute == "optimizer":
                    continue

                # "old_val" is a helper storage in compiled RecurrentTransferMechanism
                # to workaround the fact that compiled projections do no pull values
                # from their source output ports
                if attribute == "old_val":
                    continue

                # Handle PNL parameters
                pnl_param = getattr(component.parameters, attribute)

                # Use ._get to retrieve underlying numpy arrays
                # (.get will extract a scalar if originally set as a scalar)
                pnl_value = pnl_param._get(context=context)

                # Recurse if the value is a PNL object with its own parameters
                if hasattr(pnl_value, 'parameters'):
                    self._copy_params_to_pnl(context=context,
                                             component=pnl_value,
                                             params=numpy_field,
                                             ids=ids)

                elif attribute == "input_ports" or attribute == "output_ports":
                    _enumerate_recurse(pnl_value)

                # Writeback parameter value
                else:

                    # Replace empty structures with None
                    if numpy_field.nbytes == 0:
                        value = None
                    else:
                        value = numpy_field

                        # Stateful parameters include history, get the most recent value
                        if "state" in ids:
                            value = value[-1]

                        # Reshape to match the shape of the old value.
                        # Do not try to reshape ragged arrays.
                        if getattr(pnl_value, 'dtype', object) != object and pnl_value.shape != value.shape:

                            # Reshape to match numpy 0d arrays and "matrix"
                            # parameters that are flattened in compiled form
                            assert pnl_value.shape == () or pnl_param.name == "matrix", \
                                "{}: {} vs. {}".format(pnl_param.name, pnl_value.shape, value.shape)

                            # Use an assignment instead of reshape().
                            # The latter would silently create a copy if the shape
                            # could not be achieved in metadata (stride, type, ...)
                            value.shape = pnl_value.shape

                    pnl_param.set(value, context=context, override=True, compilation_sync=True)

    def _get_indexable(self, np_array):
        # outputs in recarrays need to be converted to list/tuple to be indexable
        return np_array.tolist() if np_array.dtype.base.shape == () else np_array

class CUDAExecution(Execution):
    def __init__(self, buffers=['param_struct', 'state_struct']):
        super().__init__()

        # Initialize GPU buffer map
        self._gpu_buffers = {"_" + b: None for b in buffers}

    def __get_cuda_arg(self, struct_name, arg_handler):
        gpu_buffer = self._gpu_buffers[struct_name]

        np_struct = getattr(self, struct_name)

        # .array is a public member of pycuda's In/Out ArgumentHandler classes
        if gpu_buffer is None or gpu_buffer.array is not np_struct:

            # 0-sized structures fail to upload use a dummy numpy array instead
            gpu_buffer = arg_handler(np_struct if np_struct.nbytes > 0 else np.zeros(2))

            self._gpu_buffers[struct_name] = gpu_buffer

        return gpu_buffer

    @property
    def _cuda_param_struct(self):
        return self.__get_cuda_arg("_param_struct", jit_engine.pycuda.driver.In)

    @property
    def _cuda_state_struct(self):
        return self.__get_cuda_arg("_state_struct", jit_engine.pycuda.driver.InOut)

    @property
    def _cuda_data_struct(self):
        return self.__get_cuda_arg("_data_struct", jit_engine.pycuda.driver.InOut)

    @property
    def _cuda_conditions(self):
        return self.__get_cuda_arg("_conditions", jit_engine.pycuda.driver.InOut)


class FuncExecution(CUDAExecution):

    def __init__(self, component, execution_id=None, *, tags=frozenset()):
        super().__init__()

        self._bin_func = pnlvm.LLVMBinaryFunction.from_obj(component, tags=tags)
        self._execution_context = Context(execution_id=execution_id)
        self._component = component

    @property
    def _obj(self):
        return self._component

    @property
    def _param_struct(self):
        return self._get_compilation_param('_param', '_get_param_initializer', 0)

    @property
    def _state_struct(self):
        return self._get_compilation_param('_state', '_get_state_initializer', 1)

    def execute(self, variable):
        new_variable = np.asarray(variable, dtype=self._bin_func.np_arg_dtypes[2].base)
        data_in = new_variable.reshape(self._bin_func.np_arg_dtypes[2].shape)

        data_out = self._bin_func.np_buffer_for_arg(3)

        self._bin_func(self._param_struct, self._state_struct, data_in, data_out)

        return self._get_indexable(data_out)

    def cuda_execute(self, variable):
        # Create input argument, PyCUDA doesn't care about shape
        data_in = np.asarray(variable, dtype=self._bin_func.np_arg_dtypes[2].base)
        data_out = self._bin_func.np_buffer_for_arg(3)

        self._bin_func.cuda_call(self._cuda_param_struct,
                                 self._cuda_state_struct,
                                 jit_engine.pycuda.driver.In(data_in),
                                 jit_engine.pycuda.driver.Out(data_out))

        return self._get_indexable(data_out)


class MechExecution(FuncExecution):
    pass


class CompExecution(CUDAExecution):

    active_executions = weakref.WeakSet()

    def __init__(self, composition, context:Context, *, additional_tags=frozenset()):
        super().__init__(buffers=['state_struct', 'param_struct', 'data_struct', 'conditions'])
        self._composition = composition
        self._execution_context = context
        self.__bin_exec_func = None
        self.__bin_func = None
        self.__bin_run_func = None
        self.__frozen_values = None
        self.__tags = frozenset(additional_tags)

        # Scheduling conditions, only used by "execute"
        self.__conditions = None

        self.active_executions.add(self)

    def __del__(self):
        self.active_executions.discard(self)

    @staticmethod
    def get(composition, context:Context, additional_tags=frozenset()):
        executions = composition._compilation_data.execution._get(context)
        if executions is None:
            executions = dict()
            composition._compilation_data.execution._set(executions, context)

        execution = executions.get(additional_tags, None)
        if execution is None:
            execution = pnlvm.CompExecution(composition, context, additional_tags=additional_tags)
            executions[additional_tags] = execution

        return execution

    @property
    def _obj(self):
        return self._composition

    @property
    def _bin_func(self):
        if self.__bin_func is not None:
            return self.__bin_func
        if self.__bin_exec_func is not None:
            return self.__bin_exec_func
        if self.__bin_run_func is not None:
            return self.__bin_run_func

        assert False, "Binary function not set for execution!"

    def _set_bin_node(self, node):
        assert node in self._composition._all_nodes
        node_assembly = builder_context.LLVMBuilderContext.get_current().get_node_assembly(self._composition, node)
        self.__bin_func = pnlvm.LLVMBinaryFunction.from_obj(node_assembly, tags=self.__tags.union({"node_assembly"}))

    @property
    def _conditions(self):
        if self.__conditions is None:
            gen = scheduler.ConditionGenerator(None, self._composition)

            conditions_ctype = self._bin_func.byref_arg_types[4]
            conditions_initializer = gen.get_condition_initializer()

            ct_conditions = conditions_ctype(*conditions_initializer)
            np_conditions = np.frombuffer(ct_conditions, dtype=self._bin_func.np_arg_dtypes[4], count=1)

            np_conditions.shape = ()

            self.__conditions = np_conditions

            if "stat" in self._debug_env:
                print("Instantiated condition struct ( size:" ,
                      _pretty_size(np_conditions.nbytes), ")",
                      "for", self._composition.name)

        return self.__conditions

    @property
    def _param_struct(self):
        return self._get_compilation_param('_param', '_get_param_initializer', 1)

    @property
    def _state_struct(self):
        return self._get_compilation_param('_state', '_get_state_initializer', 0)

    @property
    def _data_struct(self):
        # Run wrapper changed argument order
        arg = 2 if self._bin_func is self.__bin_run_func else 3
        return self._get_compilation_param('_data', '_get_data_initializer', arg)

    @_data_struct.setter
    def _data_struct(self, data_struct):
        self._data = data_struct

    def extract_node_struct(self, node, data):
        # state structure consists of a list of node states,
        #   followed by a list of projection contexts; get the first one
        # parameter structure consists of a list of node parameters,
        #   followed by a list of projection parameters; get the first one
        # output structure consists of a list of node outputs,
        #   followed by a list of nested data structures; get the first one
        all_nodes = data[data.dtype.names[0]]

        # Get the index into the array of all nodes
        index = self._composition._get_node_index(node)
        node_struct = all_nodes[all_nodes.dtype.names[index]]

        # Return copies of the extracted functions to avoid corrupting the
        # returned results in next execution
        return node_struct.copy().tolist() if node_struct.shape == () else node_struct.copy()

    def extract_frozen_node_output(self, node):
        return self.extract_node_struct(node, self.__frozen_values)

    def extract_node_output(self, node):
        return self.extract_node_struct(node, self._data_struct)

    def extract_node_state(self, node):
        return self.extract_node_struct(node, self._state_struct)

    def extract_node_params(self, node):
        return self.extract_node_struct(node, self._param_struct)

    def insert_node_output(self, node, data):
        # output structure consists of a list of node outputs,
        #   followed by a list of nested data structures; get the first one
        all_nodes = self._data_struct[self._data_struct.dtype.names[0]]

        # Get the index into the array of all nodes
        index = self._composition._get_node_index(node)
        value = all_nodes[all_nodes.dtype.names[index]]
        np.copyto(value, np.asarray(data, dtype=value.dtype))

    def _get_input_struct(self, inputs):
        # Either node or composition execute.

        # Read provided input data and parse into an array (generator)
        data = self._composition._build_variable_for_input_CIM(inputs)

        np_input = np.asarray(_tupleize(data), dtype=self._bin_func.np_arg_dtypes[2].base)
        np_input = np_input.reshape(self._bin_func.np_arg_dtypes[2].shape)

        if "stat" in self._debug_env:
            print("Input struct size:", _pretty_size(np_input.nbytes), "for", self._composition.name)

        return np_input

    def freeze_values(self):
        self.__frozen_values = self._data_struct.copy()

    def execute_node(self, node, inputs=None):
        # We need to reconstruct the input dictionary here if it was not provided.
        # This happens during node execution of nested compositions.
        context = self._execution_context

        if inputs is None and node is self._composition.input_CIM:

            port_inputs = {origin_port:[proj.parameters.value._get(context) for proj in p[0].path_afferents] for (origin_port, p) in self._composition.input_CIM_ports.items()}
            inputs = {}
            for p, v in port_inputs.items():
                data = inputs.setdefault(p.owner, [0] * len(p.owner.input_ports))
                index = p.owner.input_ports.index(p)
                data[index] = v[0]

        assert inputs is not None or node is not self._composition.input_CIM

        # Set bin node to make sure self._*struct works as expected
        self._set_bin_node(node)

        # Numpy doesn't allow to pass NULL to the called function.
        # Create and pass a dummy buffer filled with NaN instead.
        if inputs is not None:
            inputs = self._get_input_struct(inputs)
        else:
            inputs = self._bin_func.np_buffer_for_arg(2)

        # Nodes other than input_CIM/parameter_CIM take inputs from projections
        # and need frozen values available
        if node is not self._composition.input_CIM and node is not self._composition.parameter_CIM:
            assert self.__frozen_values is not None
            data_in = self.__frozen_values
        else:
            # The ndarray argument check doesn't allow None for null so just provide
            # the same structure as outputs.
            data_in = self._data_struct

        self._bin_func(self._state_struct, self._param_struct, inputs, data_in, self._data_struct)

        if "comp_node_debug" in self._debug_env:
            print("RAN: {}. State: {}".format(node, self.extract_node_state(node)))
            print("RAN: {}. Params: {}".format(node, self.extract_node_params(node)))
            print("RAN: {}. Results: {}".format(node, self.extract_node_output(node)))

        node._propagate_most_recent_context(context)

    @property
    def _bin_exec_func(self):
        if self.__bin_exec_func is None:
            self.__bin_exec_func = pnlvm.LLVMBinaryFunction.from_obj(self._composition, tags=self.__tags)

        return self.__bin_exec_func

    def execute(self, inputs):
        # NOTE: Make sure that input struct generation is inlined.
        # We need the binary function to be setup for it to work correctly.
        self._bin_exec_func(self._state_struct,
                            self._param_struct,
                            self._get_input_struct(inputs),
                            self._data_struct,
                            self._conditions)

    # Methods used to accelerate "Run"
    def _get_run_input_struct(self, inputs, num_input_sets, arg=3):
        # Callers that override input arg, should ensure that _bin_func is not None
        bin_f = self._bin_run_func if arg == 3 else self._bin_func

        input_type = bin_f.byref_arg_types[arg]
        c_input_type = (input_type * num_input_sets)

        # Extract input for each trial and execution id
        run_inputs = (([x] for x in self._composition._build_variable_for_input_CIM({k:v[i] for k,v in inputs.items()})) for i in range(num_input_sets))
        c_inputs = c_input_type(*_tupleize(run_inputs))
        if "stat" in self._debug_env:
            print("Instantiated struct: input ( size:" ,
                  _pretty_size(ctypes.sizeof(c_inputs)),
                  ")",
                  "for",
                  self._obj.name)

        return c_inputs

    def _get_generator_run_input_struct(self, inputs, runs):
        # Extract input for each trial
        run_inputs = ((np.atleast_2d(x) for x in self._composition._build_variable_for_input_CIM({k:np.atleast_1d(v) for k,v in inp.items()})) for inp in inputs)
        run_inputs = _tupleize(run_inputs)
        num_input_sets = len(run_inputs)
        runs = num_input_sets if runs == 0 or runs == sys.maxsize else runs
        c_input_type = self._bin_run_func.byref_arg_types[3] * num_input_sets
        return c_input_type(*run_inputs), runs

    @property
    def _bin_run_func(self):
        if self.__bin_run_func is None:
            self.__bin_run_func = pnlvm.LLVMBinaryFunction.from_obj(self._composition,
                                                                    tags=self.__tags.union({"run"}),
                                                                    ctype_ptr_args=(3,),
                                                                    dynamic_size_args=(4,))

        return self.__bin_run_func

    def _prepare_run(self, inputs, runs, num_input_sets):

        # Create input buffer
        if isgenerator(inputs):
            inputs, runs = self._get_generator_run_input_struct(inputs, runs)
            assert num_input_sets == 0 or num_input_sets == sys.maxsize
            num_input_sets = len(inputs)
        else:
            inputs = self._get_run_input_struct(inputs, num_input_sets)

        # Create output buffer
        outputs = self._bin_func.np_buffer_for_arg(4, extra_dimensions=(runs,))
        assert ctypes.sizeof(self._bin_run_func.byref_arg_types[4]) * runs == outputs.nbytes

        if "stat" in self._debug_env:
            print("Output struct size:", _pretty_size(outputs.nbytes), "for", self._composition.name)

        runs_count = np.asarray(runs, dtype=np.uint32).copy()
        input_count = np.asarray(num_input_sets, dtype=np.uint32)

        return inputs, outputs, runs_count, input_count

    def run(self, inputs, runs, num_input_sets):
        ct_inputs, outputs, runs_count, input_count = self._prepare_run(inputs, runs, num_input_sets)

        self._bin_run_func(self._state_struct,
                           self._param_struct,
                           self._data_struct,
                           ct_inputs,
                           outputs,
                           runs_count,
                           input_count)

        # Extract only #trials elements in case the run exited early
        assert runs_count <= runs, "Composition ran more times than allowed!"
        return self._get_indexable(outputs[0:runs_count])

    def cuda_run(self, inputs, runs, num_input_sets):
        ct_inputs, outputs, runs_count, input_count = self._prepare_run(inputs, runs, num_input_sets)

        self._bin_run_func.cuda_call(self._cuda_state_struct,
                                     self._cuda_param_struct,
                                     self._cuda_data_struct,
                                     jit_engine.pycuda.driver.In(np.ctypeslib.as_array(ct_inputs)),
                                     jit_engine.pycuda.driver.Out(outputs),
                                     jit_engine.pycuda.driver.InOut(runs_count),
                                     jit_engine.pycuda.driver.In(input_count))

        # Extract only #trials elements in case the run exited early
        assert runs_count <= runs, "Composition ran more times than allowed: {}".format(runs)
        return self._get_indexable(outputs[0:runs_count])

    def _prepare_evaluate(self, inputs, num_input_sets, num_evaluations, all_results:bool):
        ocm = self._composition.controller

        eval_type = "evaluate_type_all_results" if all_results else "evaluate_type_objective"
        tags = {"evaluate", "alloc_range", eval_type}
        bin_func = pnlvm.LLVMBinaryFunction.from_obj(ocm, tags=frozenset(tags), ctype_ptr_args=(5,), dynamic_size_args=(4,))
        self.__bin_func = bin_func

        # There are 8 arguments to evaluate_alloc_range:
        # comp_param, comp_state, from, to, results, input, comp_data, num_inputs
        # all but #4 are shared
        assert len(bin_func.byref_arg_types) == 8

        # Directly initialized structures
        assert ocm.agent_rep is self._composition
        comp_params = self._get_compilation_param('_eval_param', '_get_param_initializer', 0)
        comp_state = self._get_compilation_param('_eval_state', '_get_state_initializer', 1)
        comp_data = self._get_compilation_param('_eval_data', '_get_data_initializer', 6)

        # Construct input variable, the 5th parameter of the evaluate function
        ct_inputs = self._get_run_input_struct(inputs, num_input_sets, 5)

        # Output buffer
        extra_dims = (num_evaluations,)
        if all_results:
            num_trials = ocm.parameters.num_trials_per_estimate.get(self._execution_context)
            assert num_trials is not None
            extra_dims = extra_dims + (num_trials,)

        outputs = self._bin_func.np_buffer_for_arg(4, extra_dimensions=extra_dims)

        num_inputs = np.asarray(num_input_sets, dtype=np.uint32)
        if "stat" in self._debug_env:
            print("Evaluate result struct type size:",
                  _pretty_size(ctypes.sizeof(outputs.nbytes)),
                  "( evaluations:", num_evaluations, "element size:", ctypes.sizeof(out_el_ty), ")",
                  "for", self._obj.name)

        return comp_params, comp_state, comp_data, ct_inputs, outputs, num_inputs

    def cuda_evaluate(self, inputs, num_input_sets, num_evaluations, all_results:bool=False):
        comp_params, comp_state, comp_data, ct_inputs, results, num_inputs = \
            self._prepare_evaluate(inputs, num_input_sets, num_evaluations, all_results)

        cuda_args = (jit_engine.pycuda.driver.In(comp_params),
                     jit_engine.pycuda.driver.In(comp_state),
                     jit_engine.pycuda.driver.Out(results),                             # results
                     jit_engine.pycuda.driver.In(np.ctypeslib.as_array(ct_inputs)),     # inputs
                     jit_engine.pycuda.driver.In(comp_data),                            # composition data
                     jit_engine.pycuda.driver.In(num_inputs),                           # number of inputs
                    )

        self.__bin_func.cuda_call(*cuda_args, threads=int(num_evaluations))

        return results

    def thread_evaluate(self, inputs, num_input_sets, num_evaluations, all_results:bool=False):
        comp_params, comp_state, comp_data, ct_inputs, outputs, num_inputs = \
            self._prepare_evaluate(inputs, num_input_sets, num_evaluations, all_results)

        jobs = min(os.cpu_count(), num_evaluations)
        evals_per_job = (num_evaluations + jobs - 1) // jobs

        parallel_start = time.time()
        with concurrent.futures.ThreadPoolExecutor(max_workers=jobs) as ex:

            # Create input typed cast once, it is the same for every submitted job.
            input_arg = ctypes.cast(ct_inputs, self.__bin_func.c_func.argtypes[5])

            # numpy dynamic args expect only one extra dimension
            output_arg = outputs.reshape(-1, *self.__bin_func.np_arg_dtypes[4].shape)
            assert output_arg.base is outputs

            # There are 8 arguments to evaluate_alloc_range:
            # comp_param, comp_state, from, to, results, input, comp_data, input length
            results = [ex.submit(self.__bin_func,
                                 comp_params,
                                 comp_state,
                                 int(i * evals_per_job),
                                 min((i + 1) * evals_per_job, num_evaluations),
                                 output_arg,
                                 input_arg,
                                 comp_data,
                                 num_inputs)
                       for i in range(jobs)]

        parallel_stop = time.time()
        if "time_stat" in self._debug_env:
            print("Time to run {} executions of '{}' in {} threads: {}".format(
                      num_evaluations, self.__bin_func.name, jobs,
                      parallel_stop - parallel_start))


        exceptions = [r.exception() for r in results]
        assert all(e is None for e in exceptions), "Not all jobs finished sucessfully: {}".format(exceptions)

        return outputs



# from .Beukers_et_al_2022 import *
# from .Cohen_Huston1994 import *
# from .Cohen_Huston1994_horse_race import *
# from .GilzenratModel import *
# from .Kalanthroff_PCTC_2018 import *
# from .MontagueDayanSejnowski96 import *
# from .Nieuwenhuis2005Model import *



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* Bustamante_Stroop_XOR_LVOC_Model ***************************************

# along with str cast in log line 1516 vv
# next_eid_entry_data += ", \'" + "\', \'".join(str(i[0]) if isinstance(i, list) else i for i in data[0]) + "\'"

"""
Implements a model of the `Stroop XOR task
<https://scholar.google.com/scholar?hl=en&as_sdt=0%2C31&q=laura+bustamante+cohen+musslick&btnG=>`_
using a version of the `Learned Value of Control Model
<https://journals.plos.org/ploscompbiol/article?id=10.1371/journal.pcbi.1006043&rev=2>`_
"""

import numpy as np
import psyneulink as pnl

# from build_input import xor_dict
from build_stimuli_VZ import xor_dict

np.random.seed(0)


def w_fct(stim, color_control):
    """function for word_task, to modulate strength of word reading based on 1-strength of color_naming ControlSignal"""
    print('color control: ', color_control)
    return stim * (1 - color_control)


w_fct_UDF = pnl.UserDefinedFunction(custom_function=w_fct, color_control=1)


def objective_function(v):
    """function used for ObjectiveMechanism of lvoc
     v[0] = output of DDM: [probability of color naming, probability of word reading]
     v[1] = reward:        [color naming rewarded, word reading rewarded]
     """
    prob_upper = v[0]
    prob_lower = v[1]
    reward_upper = v[2][0]
    reward_lower = v[2][1]
    return prob_upper * reward_upper + prob_lower * reward_lower
    # return np.sum(v[0] * v[1])


color_stim = pnl.TransferMechanism(name='Color Stimulus', input_shapes=8)
word_stim = pnl.TransferMechanism(name='Word Stimulus', input_shapes=8)

color_task = pnl.TransferMechanism(name='Color Task')
word_task = pnl.ProcessingMechanism(name='Word Task', function=w_fct_UDF)

reward = pnl.TransferMechanism(name='Reward', input_shapes=2)

task_decision = pnl.DDM(
    name='Task Decision',
    # function=pnl.NavarroAndFuss,
    output_ports=[
        pnl.PROBABILITY_UPPER_THRESHOLD,
        pnl.PROBABILITY_LOWER_THRESHOLD
    ]
)

task_decision.set_log_conditions('func_drift_rate')
task_decision.set_log_conditions('mod_drift_rate')
task_decision.set_log_conditions('PROBABILITY_LOWER_THRESHOLD')
task_decision.set_log_conditions('PROBABILITY_UPPER_THRESHOLD')
color_task.set_log_conditions('value')
word_task.set_log_conditions('value')


c = pnl.Composition(name='Stroop XOR Model')
c.add_node(color_stim)
c.add_node(word_stim)
c.add_node(color_task, required_roles=pnl.NodeRole.ORIGIN)
c.add_node(word_task, required_roles=pnl.NodeRole.ORIGIN)
c.add_node(reward)
c.add_node(task_decision)
c.add_projection(sender=color_task, receiver=task_decision)
c.add_projection(sender=word_task, receiver=task_decision)


lvoc = pnl.OptimizationControlMechanism(
    name='LVOC ControlMechanism',
    state_features=[color_stim.input_port, word_stim.input_port],
    # state_features={pnl.SHADOW_EXTERNAL_INPUTS: [color_stim, word_stim]},
    objective_mechanism=pnl.ObjectiveMechanism(
        name='LVOC ObjectiveMechanism',
        monitor=[task_decision.output_ports[pnl.PROBABILITY_UPPER_THRESHOLD],
                 task_decision.output_ports[pnl.PROBABILITY_LOWER_THRESHOLD],
                 reward],
        # monitored_output_ports=[task_decision, reward],
        function=objective_function
    ),
    agent_rep=pnl.RegressionCFA(
        update_weights=pnl.BayesGLM(mu_0=-0.17, sigma_0=9.0909), # -0.17, 9.0909 precision = 0.11; 1/p = v
        prediction_terms=[pnl.PV.FC, pnl.PV.COST]
    ),
    function=pnl.GradientOptimization(
        convergence_criterion=pnl.VALUE,
        convergence_threshold=0.001, #0.001
        step_size=1, #1
        annealing_function=lambda x, y: x / np.sqrt(y)
        # direction=pnl.ASCENT
    ),
    control_signals=pnl.ControlSignal(
        modulates=[(pnl.SLOPE, color_task), ('color_control', word_task)],
        # function=pnl.ReLU,
        function=pnl.Logistic,
        cost_options=[pnl.CostFunctions.INTENSITY, pnl.CostFunctions.ADJUSTMENT],
        intensity_cost_function=pnl.Exponential(rate=0.25, bias=-3),
        adjustment_cost_function=pnl.Exponential(rate=0.25, bias=-3),
        # allocation_samples=[i / 2 for i in list(range(0, 50, 1))]
    )
    )

# print(lvoc.loggable_items)
lvoc.set_log_conditions('value')
# print(lvoc.loggable_items)
# lvoc.set_log_conditions('variable')
# lvoc.agent_rep.set_log_conditions('regression_weights')

lvoc.reportOutputPref=True

c.add_node(lvoc)

# c.show_graph(show_node_structure=pnl.ALL)

    # input_dict = {
    #     color_stim: [[1, 0, 0, 0, 0, 0, 0, 0]],
    #     word_stim: [[1, 0, 0, 0, 0, 0, 0, 0]],
    #     color_task: [[1]],
    #     word_task: [[-1]],
    #     reward: [[1, 0]]
    # }


# print('PREDICTION WEIGHTS T1', lvoc.agent_rep.parameters.regression_weights.get(None))

# for i in range(len(xor_dict)): # run on all 30 subjects
for i in range(3): # testing for three subjects, 200 trials per subject
    input_dict = {color_stim: xor_dict[i][0],
              word_stim: xor_dict[i][1],
              color_task: xor_dict[i][2],
              word_task: xor_dict[i][3],
              reward:    xor_dict[i][4]}

    # print("input dict: ", input_dict)

    # start_time = time.time()

    def print_weights():
        print("OUTCOME = ", lvoc.objective_mechanism.output_ports[pnl.OUTCOME].value)
        print("WEIGHTS = ", lvoc.agent_rep.parameters.regression_weights.get(i))
        print("LVOC VALUE = ", lvoc.value)
    # duration = timeit.timeit(c.run(inputs=input_dict, context=i), number=1) #number=2
    c.run(inputs=input_dict,
          context=i,
          call_after_trial=print_weights) #number=2, num_trials
    # duration = time.time() - start_time
    # print('PREDICTION WEIGHTS T2', lvoc.agent_rep.parameters.regression_weights.get(i))
    # print("WEIGHTS = ", lvoc.agent_rep.parameters.regression_weights.get(i))
    # print("OUTCOME = ", lvoc.objective_mechanism.output_ports[pnl.OUTCOME].value)
    # print('LVOC Log\n')
    # print(lvoc.log.csv)
    # print('Task Decision Log\n')
    # print(task_decision.log.csv())
    # print('Color Task Log\n')
    # print(color_task.log.csv())
    # print('Word Task Log\n')
    # print(word_task.log.csv())

    print('\n')
    print('Subject: ', i + 1)
    print('--------------------')
    print('ControlSignal variables: ', [sig.parameters.variable.get(i) for sig in lvoc.control_signals])
    print('ControlSignal values: ', [sig.parameters.value.get(i) for sig in lvoc.control_signals])
    # print('state_features: ', lvoc.get_feature_values(context=c))
    print('lvoc: ', lvoc.evaluate_agent_rep([sig.parameters.variable.get(i) for sig in lvoc.control_signals], context=i))
    # print('time: ', duration)
    print('--------------------')

# ------------------------------------------------------------------------------------------------------------------

print('\n\n\nLVOC Log\n')
print(lvoc.log.csv())
# print(lvoc.log.nparray())
# # print(lvoc.log.print_entries())

# print('\n\n\nTask Decision Log\n')
# # print(task_decision.log.csv())
# print(task_decision.log.nparray())
# print('\n\n\nColor Task Log\n')
# # print(color_task.log.csv())
# print(color_task.log.nparray())
# print('\n\n\nWord Task Log\n')
# # print(word_task.log.csv())
# print(word_task.log.nparray())

# file_lvoc = open("LVOC_200_LOG.csv", 'w')
# file_lvoc.write(lvoc.log.csv())
# file_lvoc.close()

# file_task = open("LVOC_200_TASK.csv", 'w')
# # file_task.write(task_decision.log.csv())
# file_task.close()


# def run():
#     c.run(inputs=input_dict, num_trials=1)


# duration = timeit.timeit(run, number=2)

# print('\n')
# print('--------------------')
# print('ControlSignal variables: ', [sig.parameters.variable.get(c) for sig in lvoc.control_signals])
# print('ControlSignal values: ', [sig.parameters.value.get(c) for sig in lvoc.control_signals])
# # print('state_features: ', lvoc.get_feature_values(context=c))
# print('lvoc: ', lvoc.evaluate_agent_rep([sig.parameters.variable.get(c) for sig in lvoc.control_signals], context=c))
# print('time: ', duration)
# print('--------------------')



#%%
import numpy as np
import psyneulink as pnl
import pandas as pd

from psyneulink.core.globals.utilities import set_global_seed

# # Let's make things reproducible
set_global_seed(0)

# High-level parameters the impact performance of the test
num_trials = 50
time_step_size = 0.01
num_estimates = 4000

ddm_params = dict(
    starting_value=0.0,
    rate=0.3,
    noise=1.0,
    threshold=0.6,
    non_decision_time=0.15,
    time_step_size=time_step_size,
)

# Create a simple one mechanism composition containing a DDM in integrator mode.
decision = pnl.DDM(
    function=pnl.DriftDiffusionIntegrator(**ddm_params),
    output_ports=[pnl.DECISION_OUTCOME, pnl.RESPONSE_TIME],
    name="DDM",
)

comp = pnl.Composition(pathways=decision)

# Let's generate an "experimental" dataset to fit. This is a parameter recovery test
# Lets make 10% of the trials have a positive stimulus drift rate, and the other 90%
# have a negative stimulus drift rate.
# trial_inputs = np.ones((num_trials, 1))
rng = np.random.default_rng(12345)
trial_inputs = rng.choice(
    [5.0, -5.0], size=(num_trials, 1), p=[0.10, 0.9], replace=True
)

# Make the first and last input positive for sure. This helps make sure inputs are really getting
# passed to the composition correctly during parameter fitting, and we aren't just getting a single
# trials worth of a cached input.
trial_inputs[0] = np.abs(trial_inputs[0])
trial_inputs[-1] = np.abs(trial_inputs[-1])

inputs_dict = {decision: trial_inputs}

# Store the results of this "experiment" as a numpy array. This should be a
# 2D array of shape (len(input), 2). The first column being a discrete variable
# specifying whether the upper or lower decision boundary is reached and the second column is the
# reaction time. We will put the data into a pandas DataFrame, this makes it
# easier to specify which columns in the data are categorical or not.

# Run the composition to generate some data to fit
comp.run(inputs=inputs_dict)
results = comp.results

data_to_fit = pd.DataFrame(
    np.squeeze(np.array(results)), columns=["decision", "response_time"]
)
data_to_fit["decision"] = data_to_fit["decision"].astype("category")

# Create a parameter estimation composition to fit the data we just generated and hopefully recover the
# parameters of the DDM.

fit_parameters = {
    ("rate", decision): np.linspace(-0.5, 0.5, 1000),
    ("threshold", decision): np.linspace(0.5, 1.0, 1000),
    ('non_decision_time', decision): np.linspace(0.0, 1.0, 1000),
}

#%%
import optuna

pec = pnl.ParameterEstimationComposition(
    name="pec",
    nodes=[comp],
    parameters=fit_parameters,
    outcome_variables=[
        decision.output_ports[pnl.DECISION_OUTCOME],
        decision.output_ports[pnl.RESPONSE_TIME],
    ],
    data=data_to_fit,
    optimization_function=pnl.PECOptimizationFunction(method=optuna.samplers.CmaEsSampler(seed=0), max_iterations=1000),
    num_estimates=num_estimates,
    initial_seed=42,
)

pec.controller.parameters.comp_execution_mode.set("LLVM")
pec.controller.function.parameters.save_values.set(True)
ret = pec.run(inputs={comp: trial_inputs})
optimal_parameters = list(pec.optimized_parameter_values.values())

# Check that the parameters are recovered and that the log-likelihood is correct, set the tolerance pretty high,
# things are noisy because of the low number of trials and estimates.
assert np.allclose(
    optimal_parameters,
    [ddm_params["rate"], ddm_params["threshold"], ddm_params["non_decision_time"]],
    atol=0.1,
)

records = []
for (name, mech), recovered_param in zip(fit_parameters.keys(), optimal_parameters):
    percent_error = 100.0 * (abs(ddm_params[name] - recovered_param) / ddm_params[name])
    records.append((name, mech.name, ddm_params[name], recovered_param, percent_error))
df = pd.DataFrame(records, columns=['Parameter', 'Component', 'Value', 'Recovered Value', 'Percent Error'])
print(df)



import functools
import numpy as np
import pytest

import psyneulink as pnl
import psyneulink.core.globals.keywords as kw
import psyneulink.core.llvm as pnlvm

SIZE=10
np.random.seed(0)
test_var = np.random.rand(SIZE)
test_matrix = np.random.rand(SIZE, SIZE)
test_matrix_s = np.random.rand(SIZE, SIZE // 4)
test_matrix_l = np.random.rand(SIZE, 3 * SIZE)

RAND1 = np.random.rand()
RAND2 = np.random.rand()
RAND3 = np.random.rand()
RAND4 = np.random.rand()

softmax_helper  = RAND1 * test_var
softmax_helper  = softmax_helper - np.max(softmax_helper)
softmax_helper  = np.exp(softmax_helper) / np.sum(np.exp(softmax_helper))
softmax_helper2 = np.array((softmax_helper, softmax_helper)).reshape(2, -1)

tanh_helper = (RAND1 * (test_var + RAND2 - RAND3) + RAND4)
tanh_helper = np.tanh(tanh_helper)

gaussian_helper = np.e**(-(test_var - RAND2)**2 / (2 * RAND1**2)) / np.sqrt(2 * np.pi * RAND1)
gaussian_helper = RAND3 * gaussian_helper + RAND4

relu_helper = np.maximum(RAND1 * (test_var - RAND2), RAND3 * RAND1 *(test_var - RAND2))
logistic_helper = RAND4 / (1 + np.exp(-(RAND1 * (test_var - RAND2)))) + RAND3

def gaussian_distort_helper(seed):
    state = np.random.RandomState([seed])
    # compensate for construction
    state.normal(test_var + RAND1, RAND2)
    return RAND4 * state.normal(test_var + RAND1, RAND2) + RAND3


def binomial_distort_helper(seed):
    state = np.random.RandomState([seed])
    # compensate for construction
    state.binomial(1, p=RAND1, size=len(test_var))
    return state.binomial(1, p=(1 - RAND1), size=len(test_var)) * test_var


test_data = [
    pytest.param(pnl.Linear, test_var, {kw.SLOPE:RAND1, kw.INTERCEPT:RAND2}, test_var * RAND1 + RAND2, id="LINEAR"),
    pytest.param(pnl.Exponential, test_var, {kw.SCALE:RAND1, kw.RATE:RAND2}, RAND1 * np.exp(RAND2 * test_var), id="EXPONENTIAL"),
    pytest.param(pnl.Logistic, test_var, {kw.GAIN:RAND1, kw.X_0:RAND2, kw.OFFSET:RAND3, kw.SCALE:RAND4}, logistic_helper, id="LOGISTIC"),
    pytest.param(pnl.Tanh, test_var, {kw.GAIN:RAND1, kw.BIAS:RAND2, kw.X_0:RAND3, kw.OFFSET:RAND4}, tanh_helper, id="TANH"),
    pytest.param(pnl.ReLU, test_var, {kw.GAIN:RAND1, kw.BIAS:RAND2, kw.LEAK:RAND3}, relu_helper, id="RELU"),

    # Angle doesn't have a helper using 'test_var', hardcode bopth the input and output
    pytest.param(pnl.Angle,
                 [0.5488135,  0.71518937, 0.60276338, 0.54488318, 0.4236548,
                  0.64589411, 0.43758721, 0.891773, 0.96366276, 0.38344152],
                  {},
                 [0.85314409, 0.00556188, 0.01070476, 0.0214405,  0.05559454,
                  0.08091079, 0.21657281, 0.19296643, 0.21343805, 0.92738261, 0.00483101],
                 id="ANGLE"),

    # Distort
    pytest.param(pnl.Gaussian, test_var, {kw.STANDARD_DEVIATION:RAND1, kw.BIAS:RAND2, kw.SCALE:RAND3, kw.OFFSET:RAND4}, gaussian_helper, id="GAUSSIAN"),
    pytest.param(pnl.GaussianDistort, test_var, {kw.BIAS: RAND1, kw.VARIANCE:RAND2, kw.OFFSET:RAND3, kw.SCALE:RAND4 }, gaussian_distort_helper(0), id="GAUSSIAN DISTORT GLOBAL SEED"),
    pytest.param(pnl.GaussianDistort, test_var, {kw.BIAS: RAND1, kw.VARIANCE:RAND2, kw.OFFSET:RAND3, kw.SCALE:RAND4, 'seed':0 }, gaussian_distort_helper(0), id="GAUSSIAN DISTORT"),
    pytest.param(pnl.BinomialDistort, test_var, {'seed':0, 'p':RAND1 }, binomial_distort_helper(0), id="BINOMIAL DISTORT"),

    # SoftMax 1D input
    pytest.param(pnl.SoftMax, test_var, {kw.GAIN:RAND1, kw.PER_ITEM:False}, softmax_helper, id="SOFT_MAX ALL"),
    pytest.param(pnl.SoftMax, test_var, {kw.GAIN:RAND1, kw.OUTPUT_TYPE:pnl.ARG_MAX, kw.PER_ITEM:False},
                 np.where(softmax_helper == np.max(softmax_helper), softmax_helper, 0), id="SOFT_MAX ARG_MAX"),
    pytest.param(pnl.SoftMax, test_var, {kw.GAIN:RAND1, kw.OUTPUT_TYPE:pnl.ARG_MAX_INDICATOR, kw.PER_ITEM:False},
                 np.where(softmax_helper == np.max(softmax_helper), 1, 0), id="SOFT_MAX ARG_MAX_INDICATOR"),
    pytest.param(pnl.SoftMax, test_var, {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_VAL, kw.PER_ITEM:False},
                 np.where(softmax_helper == np.max(softmax_helper), softmax_helper, 0), id="SOFT_MAX MAX_VAL"),
    pytest.param(pnl.SoftMax, test_var, {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_INDICATOR, kw.PER_ITEM:False},
                 np.where(softmax_helper == np.max(softmax_helper), 1, 0), id="SOFT_MAX MAX_INDICATOR"),
    pytest.param(pnl.SoftMax, test_var, {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.PROB, kw.PER_ITEM:False},
                 [0.0, 0.0, 0.0, 0.0, test_var[4], 0.0, 0.0, 0.0, 0.0, 0.0], id="SOFT_MAX PROB"),

    # SoftMax 2D testing per-item
    pytest.param(pnl.SoftMax, [test_var], {kw.GAIN:RAND1, kw.PER_ITEM:True}, [softmax_helper], id="SOFT_MAX ALL 2D"),
    pytest.param(pnl.SoftMax, [test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:pnl.ARG_MAX, kw.PER_ITEM:True},
                 [np.where(softmax_helper == np.max(softmax_helper), softmax_helper, 0)], id="SOFT_MAX ARG_MAX 2D"),
    pytest.param(pnl.SoftMax, [test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:pnl.ARG_MAX_INDICATOR, kw.PER_ITEM:True},
                 [np.where(softmax_helper == np.max(softmax_helper), 1, 0)], id="SOFT_MAX ARG_MAX_INDICATOR 2D"),
    pytest.param(pnl.SoftMax, [test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_VAL, kw.PER_ITEM:True},
                 [np.where(softmax_helper == np.max(softmax_helper), softmax_helper, 0)], id="SOFT_MAX MAX_VAL 2D"),
    pytest.param(pnl.SoftMax, [test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_INDICATOR, kw.PER_ITEM:True},
                 [np.where(softmax_helper == np.max(softmax_helper), 1, 0)], id="SOFT_MAX MAX_INDICATOR 2D"),
    pytest.param(pnl.SoftMax, [test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.PROB, kw.PER_ITEM:True},
                 [[0.0, 0.0, 0.0, 0.0, test_var[4], 0.0, 0.0, 0.0, 0.0, 0.0]], id="SOFT_MAX PROB 2D"),

    # SoftMax per-item with 2 elements in input
    pytest.param(pnl.SoftMax, [test_var, test_var], {kw.GAIN:RAND1, kw.PER_ITEM: True}, softmax_helper2, id="SOFT_MAX ALL PER_ITEM"),
    pytest.param(pnl.SoftMax, [test_var, test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:pnl.ARG_MAX, kw.PER_ITEM: True},
                 np.where(softmax_helper2 == np.max(softmax_helper2), softmax_helper2, 0), id="SOFT_MAX ARG_MAX PER_ITEM"),
    pytest.param(pnl.SoftMax, [test_var, test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:pnl.ARG_MAX_INDICATOR, kw.PER_ITEM: True},
                 np.where(softmax_helper2 == np.max(softmax_helper2), 1, 0), id="SOFT_MAX ARG_MAX_INDICATOR PER_ITEM"),
    pytest.param(pnl.SoftMax, [test_var, test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_VAL, kw.PER_ITEM: True},
                 np.where(softmax_helper2 == np.max(softmax_helper2), softmax_helper2, 0), id="SOFT_MAX MAX_VAL PER_ITEM"),
    pytest.param(pnl.SoftMax, [test_var, test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_INDICATOR, kw.PER_ITEM: True},
                 np.where(softmax_helper2 == np.max(softmax_helper2), 1, 0), id="SOFT_MAX MAX_INDICATOR PER_ITEM"),

    # Linear Matrix
    pytest.param(pnl.MatrixTransform, test_var, {kw.MATRIX:test_matrix}, np.dot(test_var, test_matrix), id="LINEAR_MATRIX SQUARE"),
    pytest.param(pnl.MatrixTransform, test_var, {kw.MATRIX:test_matrix_l}, np.dot(test_var, test_matrix_l), id="LINEAR_MATRIX WIDE"),
    pytest.param(pnl.MatrixTransform, test_var, {kw.MATRIX:test_matrix_s}, np.dot(test_var, test_matrix_s), id="LINEAR_MATRIX TALL"),

    # Dropout is just identity in non-learning mode
    pytest.param(pnl.Dropout, test_var, {}, test_var, id="DROPOUT"),
]

@pytest.mark.function
@pytest.mark.transfer_function
@pytest.mark.benchmark
@pytest.mark.parametrize("func, variable, params, expected", test_data)
def test_execute(func, variable, params, expected, benchmark, func_mode):
    benchmark.group = "TransferFunction " + func.componentName

    if func_mode != 'Python':
        if ('output' in params
                and params['output'] in {kw.MAX_VAL, kw.MAX_ABS_VAL, kw.MAX_INDICATOR, kw.MAX_ABS_INDICATOR,
                                         kw.MIN_VAL, kw.MIN_ABS_VAL, kw.MIN_INDICATOR, kw.MIN_ABS_INDICATOR}):
            pytest.skip("{params['mode']} is not supported in {func_mode}")

    f = func(default_variable=variable, **params)
    ex = pytest.helpers.get_func_execution(f, func_mode)

    res = benchmark(ex, variable)
    np.testing.assert_allclose(res, expected, rtol=1e-5, atol=1e-8)


tanh_derivative_helper = (RAND1 * (test_var + RAND2) + RAND3)
tanh_derivative_helper = (1 - np.tanh(tanh_derivative_helper)**2) * RAND4 * RAND1


derivative_test_data = [
    (pnl.Linear, test_var, {kw.SLOPE:RAND1, kw.INTERCEPT:RAND2}, RAND1),
    (pnl.Exponential, test_var, {kw.SCALE:RAND1, kw.RATE:RAND2}, RAND1 * RAND2 * np.exp(RAND2 * test_var)),
    (pnl.Logistic, test_var, {kw.GAIN:RAND1, kw.X_0:RAND2, kw.OFFSET:RAND3, kw.SCALE:RAND4}, RAND1 * RAND4 * logistic_helper * (1 - logistic_helper)),
    (pnl.ReLU, test_var, {kw.GAIN:RAND1, kw.BIAS:RAND2, kw.LEAK:RAND3}, np.where((test_var - RAND2) > 0, RAND1, RAND1 * RAND3)),
    (pnl.Tanh, test_var, {kw.GAIN:RAND1, kw.BIAS:RAND2, kw.OFFSET:RAND3, kw.SCALE:RAND4}, tanh_derivative_helper),

    # SoftMax per-item=False
    (pnl.SoftMax, test_var, {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_VAL, kw.PER_ITEM:False},
     [-0.010680386821751537, -0.011118109698906909, -0.01082040340318878, -0.010670257514724047, -0.010362498859374309,
      -0.010933660158663306, -0.010397412260182806, -0.011602329078808718, 0.09684744183944892, -0.010262384043848513]),
    (pnl.SoftMax, test_var, {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_INDICATOR, kw.PER_ITEM:False},
     [-0.010680386821751537, -0.011118109698906909, -0.01082040340318878, -0.010670257514724047, -0.010362498859374309,
      -0.010933660158663306, -0.010397412260182806, -0.011602329078808718, 0.09684744183944892, -0.010262384043848513]),
    pytest.param(pnl.SoftMax, test_var, {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.ALL, kw.PER_ITEM:False},
                 [[ 0.088635686173821480, -0.010058549286956951, -0.009789214523259433, -0.009653377599514660, -0.009374948470179183,
                   -0.009891677863509920, -0.009406534609578588, -0.010496622361458180, -0.010680386821751540, -0.009284374637613039],
                  [-0.010058549286956951,  0.091856076128865180, -0.010190413769852785, -0.010049009732287338, -0.009759169518165271,
                   -0.010297076447528582, -0.009792050177702091, -0.010926813872042194, -0.011118109698906910, -0.009664883625423075],
                  [-0.009789214523259433, -0.010190413769852785,  0.089669339130699100, -0.009779930406389987, -0.009497851156931268,
                   -0.010021354713444461, -0.009529851380888969, -0.010634229847424508, -0.010820403403188785, -0.009406089929318929],
                  [-0.009653377599514660, -0.010049009732287338, -0.009779930406389987,  0.088560779144081720, -0.009366057244326959,
                   -0.009882296570138368, -0.009397613427348460, -0.010486667337129447, -0.010670257514724050, -0.009275569312222474],
                  [-0.009374948470179183, -0.009759169518165271, -0.009497851156931268, -0.009366057244326959,  0.08627659236704915,
                   -0.009597264807784339, -0.009126561218167337, -0.010184203911638403, -0.010362498859374313, -0.009008037180482098],
                  [-0.009891677863509920, -0.010297076447528582, -0.010021354713444461, -0.009882296570138368, -0.009597264807784339,
                    0.090503011588098000, -0.009629599976882700, -0.010745537931292683, -0.010933660158663310, -0.009504543118853646],
                  [-0.009406534609578588, -0.009792050177702091, -0.009529851380888969, -0.009397613427348460, -0.009126561218167337,
                   -0.009629599976882700,  0.086536526770559590, -0.010218516599910580, -0.010397412260182810, -0.009038387119898062],
                  [-0.010496622361458180, -0.010926813872042194, -0.010634229847424508, -0.010486667337129447, -0.010184203911638403,
                   -0.010745537931292683, -0.010218516599910580,  0.095380732590004670, -0.011602329078808723, -0.01008581165029997],
                  [-0.010680386821751540, -0.011118109698906910, -0.010820403403188785, -0.010670257514724050, -0.010362498859374313,
                   -0.010933660158663310, -0.010397412260182810, -0.011602329078808723,  0.096847441839448930, -0.010262384043848514],
                  [-0.009284374637613039, -0.009664883625423075, -0.009406089929318929, -0.009275569312222474, -0.009008037180482098,
                   -0.009504543118853646, -0.009038387119898062, -0.010085811650299970, -0.010262384043848514,  0.08553008061795979]],
                 marks=pytest.mark.llvm_not_implemented),
    # SoftMax per-tem=True
    (pnl.SoftMax, [test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_VAL, kw.PER_ITEM:True},
     [[-0.010680386821751537, -0.011118109698906909, -0.01082040340318878, -0.010670257514724047, -0.010362498859374309,
       -0.010933660158663306, -0.010397412260182806, -0.011602329078808718, 0.09684744183944892, -0.010262384043848513]]),
    (pnl.SoftMax, [test_var, test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_INDICATOR, kw.PER_ITEM:True},
     [[-0.010680386821751537, -0.011118109698906909, -0.01082040340318878, -0.010670257514724047, -0.010362498859374309,
       -0.010933660158663306, -0.010397412260182806, -0.011602329078808718, 0.09684744183944892, -0.010262384043848513],
      [-0.010680386821751537, -0.011118109698906909, -0.01082040340318878, -0.010670257514724047, -0.010362498859374309,
       -0.010933660158663306, -0.010397412260182806, -0.011602329078808718, 0.09684744183944892, -0.010262384043848513]]),
    pytest.param(pnl.SoftMax, [test_var], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.ALL, kw.PER_ITEM:True},
                 [[ 0.088635686173821480, -0.010058549286956951, -0.009789214523259433, -0.009653377599514660, -0.009374948470179183,
                    -0.009891677863509920, -0.009406534609578588, -0.010496622361458180, -0.010680386821751540, -0.009284374637613039],
                   [-0.010058549286956951,  0.091856076128865180, -0.010190413769852785, -0.010049009732287338, -0.009759169518165271,
                    -0.010297076447528582, -0.009792050177702091, -0.010926813872042194, -0.011118109698906910, -0.009664883625423075],
                   [-0.009789214523259433, -0.010190413769852785,  0.089669339130699100, -0.009779930406389987, -0.009497851156931268,
                    -0.010021354713444461, -0.009529851380888969, -0.010634229847424508, -0.010820403403188785, -0.009406089929318929],
                   [-0.009653377599514660, -0.010049009732287338, -0.009779930406389987,  0.088560779144081720, -0.009366057244326959,
                    -0.009882296570138368, -0.009397613427348460, -0.010486667337129447, -0.010670257514724050, -0.009275569312222474],
                   [-0.009374948470179183, -0.009759169518165271, -0.009497851156931268, -0.009366057244326959,  0.08627659236704915,
                    -0.009597264807784339, -0.009126561218167337, -0.010184203911638403, -0.010362498859374313, -0.009008037180482098],
                   [-0.009891677863509920, -0.010297076447528582, -0.010021354713444461, -0.009882296570138368, -0.009597264807784339,
                     0.090503011588098000, -0.009629599976882700, -0.010745537931292683, -0.010933660158663310, -0.009504543118853646],
                   [-0.009406534609578588, -0.009792050177702091, -0.009529851380888969, -0.009397613427348460, -0.009126561218167337,
                    -0.009629599976882700,  0.086536526770559590, -0.010218516599910580, -0.010397412260182810, -0.009038387119898062],
                   [-0.010496622361458180, -0.010926813872042194, -0.010634229847424508, -0.010486667337129447, -0.010184203911638403,
                    -0.010745537931292683, -0.010218516599910580,  0.095380732590004670, -0.011602329078808723, -0.01008581165029997],
                   [-0.010680386821751540, -0.011118109698906910, -0.010820403403188785, -0.010670257514724050, -0.010362498859374313,
                    -0.010933660158663310, -0.010397412260182810, -0.011602329078808723,  0.096847441839448930, -0.010262384043848514],
                   [-0.009284374637613039, -0.009664883625423075, -0.009406089929318929, -0.009275569312222474, -0.009008037180482098,
                    -0.009504543118853646, -0.009038387119898062, -0.010085811650299970, -0.010262384043848514,  0.08553008061795979]],
                 marks=pytest.mark.llvm_not_implemented),
]

@pytest.mark.function
@pytest.mark.transfer_function
@pytest.mark.benchmark
@pytest.mark.parametrize("func, variable, params, expected",
                         derivative_test_data,
                         ids=lambda x: getattr(x, 'name', None) or getattr(x, 'get', lambda p, q: None)(kw.OUTPUT_TYPE, None))
def test_transfer_derivative(func, variable, params, expected, benchmark, func_mode):
    benchmark.group = "TransferFunction " + func.componentName + " Derivative"

    f = func(default_variable=variable, **params)

    if func_mode == 'Python':
        ex = f.derivative

    elif func_mode == 'LLVM':
        ex = pnlvm.execution.FuncExecution(f, tags=frozenset({"derivative"})).execute

    elif func_mode == 'PTX':
        ex = pnlvm.execution.FuncExecution(f, tags=frozenset({"derivative"})).cuda_execute

    else:
        assert False, "unknown function mode: {}".format(func_mode)

    res = benchmark(ex, variable)

    # Tanh and Logistic need reduced accuracy in single precision mode
    if func_mode != 'Python' and pytest.helpers.llvm_current_fp_precision() == 'fp32' and func in {pnl.Tanh, pnl.Logistic}:
        tolerance = {'rtol': 5e-7, 'atol': 1e-8}
    else:
        tolerance = {}

    np.testing.assert_allclose(res, expected, **tolerance)


derivative_out_test_data = [
    (pnl.Logistic, logistic_helper, {kw.GAIN:RAND1, kw.X_0:RAND2, kw.OFFSET:RAND3, kw.SCALE:RAND4}, RAND1 * RAND4 * logistic_helper * (1 - logistic_helper)),
    (pnl.ReLU, relu_helper, {kw.GAIN:RAND1, kw.BIAS:RAND2, kw.LEAK:RAND3}, np.where((test_var - RAND2) > 0, RAND1, RAND1 * RAND3)),
    (pnl.SoftMax, softmax_helper, {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_VAL, kw.PER_ITEM:False},
     [-0.010680386821751537, -0.011118109698906909, -0.01082040340318878, -0.010670257514724047, -0.010362498859374309,
      -0.010933660158663306, -0.010397412260182806, -0.011602329078808718, 0.09684744183944892, -0.010262384043848513]),
    (pnl.SoftMax, [softmax_helper, softmax_helper], {kw.GAIN:RAND1, kw.OUTPUT_TYPE:kw.MAX_VAL, kw.PER_ITEM:True},
     [[-0.010680386821751537, -0.011118109698906909, -0.01082040340318878, -0.010670257514724047, -0.010362498859374309,
       -0.010933660158663306, -0.010397412260182806, -0.011602329078808718, 0.09684744183944892, -0.010262384043848513],
      [-0.010680386821751537, -0.011118109698906909, -0.01082040340318878, -0.010670257514724047, -0.010362498859374309,
       -0.010933660158663306, -0.010397412260182806, -0.011602329078808718, 0.09684744183944892, -0.010262384043848513]]),
]
@pytest.mark.function
@pytest.mark.transfer_function
@pytest.mark.benchmark
@pytest.mark.parametrize("func, variable, params, expected", derivative_out_test_data, ids=lambda x: getattr(x, 'name', None) or getattr(x, 'get', lambda p, q: None)(kw.OUTPUT_TYPE, None))
def test_transfer_derivative_out(func, variable, params, expected, benchmark, func_mode):
    benchmark.group = "TransferFunction " + func.componentName + " Derivative"

    f = func(default_variable=variable, **params)

    if func_mode == 'Python':
        def ex(x):
            return f.derivative(input=None, output=x)

    elif func_mode == 'LLVM':
        ex = pnlvm.execution.FuncExecution(f, tags=frozenset({"derivative_out"})).execute

    elif func_mode == 'PTX':
        ex = pnlvm.execution.FuncExecution(f, tags=frozenset({"derivative_out"})).cuda_execute

    else:
        assert False, "unknown function mode: {}".format(func_mode)

    res = benchmark(ex, variable)

    # Logistic needs reduced accuracy in single precision mode because it uses exp()
    if func_mode != 'Python' and func is pnl.Logistic and pytest.helpers.llvm_current_fp_precision() == 'fp32' and func is pnl.Logistic:
        tolerance = {'rtol': 1e-7, 'atol': 1e-8}
    else:
        tolerance = {}

    np.testing.assert_allclose(res, expected, **tolerance)


def combine_costs(costs):
    return functools.reduce(lambda x, y: x | y, costs, pnl.CostFunctions.NONE)

@pytest.mark.parametrize("cost_functions",
                         map(combine_costs, pytest.helpers.power_set(cf for cf in pnl.CostFunctions if cf != pnl.CostFunctions.NONE and cf != pnl.CostFunctions.ALL)))
@pytest.mark.benchmark
@pytest.mark.function
def test_transfer_with_costs(cost_functions, func_mode, benchmark):

    f = pnl.TransferWithCosts(enabled_cost_functions=cost_functions)

    def check(cost_function, if_enabled, if_disabled, observed):
        if cost_function in cost_functions:
            np.testing.assert_allclose(observed, if_enabled)
        else:
            assert np.array_equal(observed, if_disabled)

            # HACK: workaround intensity cost returning [1] when disabled
            nonlocal total_cost
            total_cost -= observed or 0

    ex = pytest.helpers.get_func_execution(f, func_mode)

    res = ex(10)
    total_cost = (f.intensity_cost or 0) + (f.adjustment_cost or 0) + (f.duration_cost or 0)

    assert res == [10]

    # TODO : Intensity cost is [1] even when disabled
    # https://github.com/PrincetonUniversity/PsyNeuLink/issues/2711
    check(pnl.CostFunctions.INTENSITY,
          [22026.465794806703],
          None if cost_functions == pnl.CostFunctions.NONE else [1],
          f.intensity_cost)
    check(pnl.CostFunctions.ADJUSTMENT, [10], None, f.adjustment_cost)
    check(pnl.CostFunctions.DURATION, [10], None, f.duration_cost)

    # TODO: Combined costs are not supported in compiled mode
    # https://github.com/PrincetonUniversity/PsyNeuLink/issues/2712
    if func_mode == "Python":
        assert np.array_equal(total_cost, f.combined_costs or 0)


    # Second run with positive adjustment
    res = ex(15)
    total_cost = (f.intensity_cost or 0) + (f.adjustment_cost or 0) + (f.duration_cost or 0)

    assert res == [15]

    # TODO : Intensity cost is [1] even when disabled
    # https://github.com/PrincetonUniversity/PsyNeuLink/issues/2711
    check(pnl.CostFunctions.INTENSITY,
          [3269017.372472108],
          None if cost_functions == pnl.CostFunctions.NONE else [1],
          f.intensity_cost)
    check(pnl.CostFunctions.ADJUSTMENT, [5], None, f.adjustment_cost)
    check(pnl.CostFunctions.DURATION, [25], None, f.duration_cost)

    # TODO: Combined costs are not supported in compiled mode
    # https://github.com/PrincetonUniversity/PsyNeuLink/issues/2712
    if func_mode == "Python":
        assert np.array_equal(total_cost, f.combined_costs or 0)


    # Third run with negative adjustment
    res = ex(7)
    total_cost = (f.intensity_cost or 0) + (f.adjustment_cost or 0) + (f.duration_cost or 0)

    assert res == [7]

    # TODO : Intensity cost is [1] even when disabled
    # https://github.com/PrincetonUniversity/PsyNeuLink/issues/2711
    check(pnl.CostFunctions.INTENSITY,
          [1096.6331584284583],
          None if cost_functions == pnl.CostFunctions.NONE else [1],
          f.intensity_cost)
    check(pnl.CostFunctions.ADJUSTMENT, [8], None, f.adjustment_cost)
    check(pnl.CostFunctions.DURATION, [32], None, f.duration_cost)

    # TODO: Combined costs are not supported in compiled mode
    # https://github.com/PrincetonUniversity/PsyNeuLink/issues/2712
    if func_mode == "Python":
        assert np.array_equal(total_cost, f.combined_costs or 0)

    benchmark(ex, 10)


def test_transfer_with_costs_toggle():
    f = pnl.TransferWithCosts()
    result = f(1)
    np.testing.assert_allclose(result, 1)
    f.toggle_cost(pnl.CostFunctions.INTENSITY)

    f = pnl.TransferWithCosts(enabled_cost_functions=pnl.CostFunctions.INTENSITY)
    result = f(2)
    np.testing.assert_allclose(result, 2)
    np.testing.assert_allclose(f.intensity_cost, 7.38905609893065)
    assert f.adjustment_cost is None
    assert f.duration_cost is None
    np.testing.assert_allclose(f.combined_costs, 7.38905609893065)

    f.toggle_cost(pnl.CostFunctions.ADJUSTMENT)
    result = f(3)
    np.testing.assert_allclose(result, 3)
    np.testing.assert_allclose(f.intensity_cost, 20.085536923187668)
    np.testing.assert_allclose(f.adjustment_cost, 1)
    assert f.duration_cost is None
    np.testing.assert_allclose(f.combined_costs, 21.085536923187668)

    f.toggle_cost(pnl.CostFunctions.DURATION)
    result = f(5)
    np.testing.assert_allclose(result, 5)
    np.testing.assert_allclose(f.intensity_cost, 148.413159102576603)
    np.testing.assert_allclose(f.adjustment_cost, 2)
    np.testing.assert_allclose(f.duration_cost, 5)
    np.testing.assert_allclose(f.combined_costs, 155.413159102576603)
    result = f(1)
    np.testing.assert_allclose(result, 1)
    np.testing.assert_allclose(f.intensity_cost, 2.718281828459045)
    np.testing.assert_allclose(f.adjustment_cost, 4)
    np.testing.assert_allclose(f.duration_cost, 6)
    np.testing.assert_allclose(f.combined_costs, 12.718281828459045)


@pytest.mark.parametrize(
    'default_variable, func_name, expected_func_variable, expected_func_value',
    [
        ([1, 2, 3], 'transfer_fct', [1, 2, 3], [1, 2, 3])
    ]
)
def test_transfer_with_costs_shapes(
    default_variable,
    func_name,
    expected_func_variable,
    expected_func_value
):
    twc = pnl.TransferWithCosts(default_variable=default_variable)

    np.testing.assert_array_equal(
        getattr(twc.parameters, func_name).get().defaults.variable,
        expected_func_variable
    )
    np.testing.assert_array_equal(
        getattr(twc.parameters, func_name).get().defaults.value,
        expected_func_value
    )



# %%
import numpy as np
import psyneulink as pnl

import matplotlib.pyplot as plt

plt.rcParams["figure.figsize"] = (20, 10)

import seaborn as sns
import pandas as pd

import wfpt

from psyneulink.core.components.functions.nonstateful.fitfunctions import simulation_likelihood


def ddm_pdf_analytical(rate, threshold, noise, starting_value, non_decision_time, time_step_size=0.01):
    from ddm import Model
    from ddm.models import DriftConstant, NoiseConstant, BoundConstant, OverlayNonDecision, ICPoint
    from ddm.functions import display_model

    model = Model(name='Simple model',
                  drift=DriftConstant(drift=rate),
                  noise=NoiseConstant(noise=noise),
                  bound=BoundConstant(B=threshold),
                  IC=ICPoint(x0=starting_value),
                  overlay=OverlayNonDecision(nondectime=non_decision_time),
                  dx=.001, dt=time_step_size, T_dur=3)
    display_model(model)
    s = model.solve()

    return model.t_domain(), s.pdf_corr(), s.pdf_err()


def ddm_pdf_simulate(drift_rate=0.75, threshold=1.0, noise=0.1, starting_point=0.0, non_decision_time=0.0,
                     time_step_size=0.01, num_samples=1000000, use_pnl=True, rt_space=None):
    if use_pnl:
        decision = pnl.DDM(function=pnl.DriftDiffusionIntegrator(starting_value=0.1234),
                           output_ports=[pnl.DECISION_VARIABLE, pnl.RESPONSE_TIME],
                           name='DDM')

        comp = pnl.Composition()
        comp.add_node(decision)

        context = pnl.Context()
        decision.function.parameters.rate.set(drift_rate, context)
        decision.function.parameters.noise.set(noise, context)
        decision.function.parameters.threshold.set(threshold, context)
        decision.function.parameters.time_step_size.set(time_step_size, context)
        decision.function.parameters.starting_value.set(np.array([starting_point]), context)
        decision.function.parameters.time_step_size.set(time_step_size, context)
        decision.function.parameters.non_decision_time.set(non_decision_time, context)

        input = np.ones((1, 1))

        comp.run(inputs={decision: input},
                 num_trials=num_samples * len(input),
                 execution_mode=pnl.ExecutionMode.LLVMRun,
                 context=context)

        results = np.squeeze(np.array(comp.results))
        rts = np.array(np.vsplit(results, len(input)))

    else:
        rts = wfpt.simulate_wfpt(starting_point * np.ones(num_samples),
                                 non_decision_time * np.ones(num_samples),
                                 drift_rate * np.ones(num_samples),
                                 threshold * np.ones(num_samples),
                                 dt=time_step_size)
        rts = np.expand_dims(np.column_stack((np.sign(rts) * threshold, np.abs(rts))), axis=0)

    # Make a histogram
    # hist = bh.Histogram(bh.axis.Boolean(), bh.axis.Regular(int(3 / time_step_size), 0.0, 3.0))
    # hist.fill(rts[:, :, 0].flatten() > 0.0, rts[:, :, 1].flatten())

    if rt_space is None:
        rt_space = np.linspace(0.0, 3.0, 3000)

    df = pd.DataFrame(index=rt_space)

    df[f'Correct KDE (dt={time_step_size})'] = simulation_likelihood(rts,
                                                                     categorical_dims=np.array([True, False]),
                                                                     combine_trials=True,
                                                                     exp_data=np.c_[
                                                                         threshold * np.ones(len(rt_space)), rt_space])
    df[f'Error KDE (dt={time_step_size})'] = simulation_likelihood(rts,
                                                                   categorical_dims=np.array([True, False]),
                                                                   combine_trials=True,
                                                                   exp_data=np.c_[
                                                                       -threshold * np.ones(len(rt_space)), rt_space])

    # df[f'Correct Histogram (dt={time_step_size})'] = (hist[True, :] / hist.sum(flow=True) / time_step_size).view()
    # df[f'Error Histogram (dt={time_step_size})'] = (hist[False, :] / hist.sum(flow=True) / time_step_size).view()

    return df


def ddm_plot_check():
    ddm_params = dict(starting_point=0.1, drift_rate=0.3, noise=1.0, threshold=0.6, non_decision_time=0.8)

    # from numpy.random import rand
    # pd.DataFrame({
    #     'drift_rate': (rand() - .5) * 8,
    #     'non_decision_time': 0.2 + rand() * 0.3,
    #     'threshold': 0.5 + rand() * 1.5,
    #     'noise': 1.0
    # })

    NUM_SAMPLES = 100000

    rt_space = np.linspace(0.0, 3.0, 30000)

    # Get the analytical
    t_domain, pdf_corr, pdf_err = ddm_pdf_analytical(**ddm_params)

    # Interpolate to common rt space
    from scipy.interpolate import interpn
    anal_df = pd.DataFrame(index=rt_space)
    anal_df[f"Correct Analytical"] = interpn((t_domain,), pdf_corr, rt_space,
                                             method='linear', bounds_error=False, fill_value=1e-10)
    anal_df[f"Error Analytical"] = interpn((t_domain,), pdf_err, rt_space,
                                           method='linear', bounds_error=False, fill_value=1e-10)

    # Navarro and Fuss solution
    # p_err = np.array(
    #     [wfpt.wfpt_logp(t, 0, starting_point, non_decision_time, drift_rate, threshold, eps=1e-10) for t in
    #      model_t_domain[1:]])
    # p_corr = np.array(
    #     [wfpt.wfpt_logp(t, 1, starting_point, non_decision_time, drift_rate, threshold, eps=1e-10) for t in
    #      model_t_domain[1:]])

    df = pd.concat([
        anal_df,
        ddm_pdf_simulate(**ddm_params, time_step_size=0.01, use_pnl=True, num_samples=NUM_SAMPLES, rt_space=rt_space),
        ddm_pdf_simulate(**ddm_params, time_step_size=0.001, use_pnl=True, num_samples=NUM_SAMPLES, rt_space=rt_space),
        ddm_pdf_simulate(**ddm_params, time_step_size=0.0001, use_pnl=True, num_samples=NUM_SAMPLES, rt_space=rt_space),
    ])

    fig, axes = plt.subplots(1, 2, sharex=True, sharey=True)

    # df = df.loc[:, ~df.columns.str.contains('Histogram')]
    sns.lineplot(data=df.filter(regex='Correct'), ax=axes[0])
    sns.lineplot(data=df.filter(regex='Error'), ax=axes[1])
    plt.show()

    plt.savefig(f"{'_'.join([f'{p}={v}' for p, v in ddm_params.items()])}.png")


ddm_plot_check()


def plot_sim_results(rate, threshold, noise, starting_value, non_decision_time, time_step_size=0.001,
                     sim_data=None, rt_space=None):
    # Get the analytical DDM results
    from pyddm import Model
    from pyddm.models import DriftConstant, NoiseConstant, BoundConstant, OverlayNonDecision, ICPoint

    model = Model(name='Simple model',
                  drift=DriftConstant(drift=rate),
                  noise=NoiseConstant(noise=noise),
                  bound=BoundConstant(B=threshold),
                  IC=ICPoint(x0=starting_value),
                  overlay=OverlayNonDecision(nondectime=non_decision_time),
                  dx=.001, dt=time_step_size, T_dur=3)
    s = model.solve()

    t_domain, pdf_corr, pdf_err = model.t_domain(), s.pdf_corr(), s.pdf_err()

    # Interpolate to common rt space
    if rt_space is None:
        rt_space = np.linspace(0.0, 3.0, 3000)

    from scipy.interpolate import interpn
    df = pd.DataFrame(index=rt_space)
    df[f"Correct Analytical"] = interpn((t_domain,), pdf_corr, rt_space,
                                        method='linear', bounds_error=False, fill_value=1e-10)
    df[f"Error Analytical"] = interpn((t_domain,), pdf_err, rt_space,
                                      method='linear', bounds_error=False, fill_value=1e-10)

    if sim_data is not None:
        sim_df = pd.DataFrame(index=rt_space)
        sim_df[f'Correct KDE'] = simulation_likelihood(sim_data,
                                                       categorical_dims=np.array([True, False]),
                                                       combine_trials=True,
                                                       exp_data=np.c_[
                                                           threshold * np.ones(len(rt_space)), rt_space])
        sim_df[f'Error KDE'] = simulation_likelihood(sim_data,
                                                     categorical_dims=np.array([True, False]),
                                                     combine_trials=True,
                                                     exp_data=np.c_[
                                                         -threshold * np.ones(len(rt_space)), rt_space])

        df = pd.concat([df, sim_df])

    fig, axes = plt.subplots(1, 2, sharex=True, sharey=True)

    # df = df.loc[:, ~df.columns.str.contains('Histogram')]
    sns.lineplot(data=df.filter(regex='Correct'), ax=axes[0])
    sns.lineplot(data=df.filter(regex='Error'), ax=axes[1])
    plt.show()



import numpy as np
import psyneulink as pnl
import pytest


class TestOutputPorts:

    @pytest.mark.mechanism
    def test_output_port_variable_spec(self, mech_mode):
        # Test specification of OutputPort's variable
        mech = pnl.ProcessingMechanism(default_variable=[[1.],[2.],[3.]],
                                       name='MyMech',
                                       output_ports=[
                                           pnl.OutputPort(name='z', variable=(pnl.OWNER_VALUE, 2)),
                                           pnl.OutputPort(name='y', variable=(pnl.OWNER_VALUE, 1)),
                                           pnl.OutputPort(name='x', variable=(pnl.OWNER_VALUE, 0)),
                                           pnl.OutputPort(name='all', variable=(pnl.OWNER_VALUE)),
                                           pnl.OutputPort(name='execution count', variable=(pnl.OWNER_EXECUTION_COUNT))
                                       ])
        expected = [[3.],[2.],[1.],[[1.],[2.],[3.]], [0]]
        for i, e in zip(mech.output_values, expected):
            assert np.array_equal(i, e)
        EX = pytest.helpers.get_mech_execution(mech, mech_mode)

        EX([[1.],[2.],[3.]])
        EX([[1.],[2.],[3.]])
        EX([[1.],[2.],[3.]])
        res = EX([[1.],[2.],[3.]])
        expected = [[3.],[2.],[1.],[[1.],[2.],[3.]], [4]]
        for i, e in zip(res, expected):
            assert np.array_equal(i, e)

    @pytest.mark.composition
    @pytest.mark.mechanism
    @pytest.mark.parametrize('spec, expected1, expected2',
                             [((pnl.OWNER_VALUE, 0), [1], [1]),
                              ((pnl.OWNER_VALUE, 1), [2], [2]),
                              ((pnl.OWNER_VALUE, 2), [3], [3]),
                              pytest.param((pnl.OWNER_VALUE, 3), [3], [3], marks=[pytest.mark.xfail(raises=IndexError, match="list index out of range")]),
                              ((pnl.OWNER_EXECUTION_COUNT), [4], [8]),
                              (("num_executions", pnl.TimeScale.LIFE), [4], [8]),
                              (("num_executions", pnl.TimeScale.RUN), [4], [4]),
                              (("num_executions", pnl.TimeScale.TRIAL), [2], [2]),
                              (("num_executions", pnl.TimeScale.PASS), [1], [1]),
                              (("num_executions", pnl.TimeScale.TIME_STEP), [1], [1]),
                             ], ids=lambda x: str(x) if len(x) != 1 else '')
    @pytest.mark.usefixtures("comp_mode_no_llvm")
    def tests_output_port_variable_spec_composition(self, comp_mode, spec, expected1, expected2):
        if (len(spec) == 2) and (spec[1] == pnl.TimeScale.RUN) and \
           ((comp_mode & pnl.ExecutionMode._Exec) == pnl.ExecutionMode._Exec):
            pytest.skip("{} is not supported in {}".format(spec[1], comp_mode))

        # Test specification of OutputPort's variable
        # OutputPort mech.output_ports['all'] has a different dimensionality than the other OutputPorts;
        #    as a consequence, when added as a terminal node, the Composition can't construct an IDENTITY_MATRIX
        #    from the mech's OutputPorts to the Composition's output_CIM.
        # FIX: Remove the following line and correct assertions below once above condition is resolved
        var = [[1], [2], [3]]
        mech = pnl.ProcessingMechanism(default_variable=var, name='MyMech',
                                       output_ports=[pnl.OutputPort(variable=spec)])
        C = pnl.Composition(name='MyComp')
        C.add_node(node=mech)
        C.termination_processing[pnl.TimeScale.TRIAL] = pnl.AtPass(2)
        outs = C.run(inputs={mech: var}, num_trials=2, execution_mode=comp_mode)
        # outs is entire mechanism value, expected is output port value
        np.testing.assert_allclose(outs[0], expected1)
        outs = C.run(inputs={mech: var}, num_trials=2, execution_mode=comp_mode)
        np.testing.assert_allclose(outs[0], expected2)

    def test_no_path_afferents(self):
        A = pnl.OutputPort()
        with pytest.raises(pnl.PortError) as error:
            A.path_afferents
        assert 'OutputPorts do not have \'path_afferents\'; (access attempted for Deferred Init OutputPort).' \
               in str(error.value)
        with pytest.raises(pnl.PortError) as error:
            A.path_afferents = ['test']
        assert 'OutputPorts are not allowed to have \'path_afferents\' ' \
               '(assignment attempted for Deferred Init OutputPort).' in str(error.value)



import functools
import numpy as np
import psyneulink as pnl
import pytest

import psyneulink.core.components.functions.nonstateful.transferfunctions

class TestLCControlMechanism:

    @pytest.mark.mechanism
    @pytest.mark.control_mechanism
    @pytest.mark.composition
    @pytest.mark.benchmark(group="LCControlMechanism Default")
    def test_lc_control_mechanism_as_controller(self, benchmark):
        G = 1.0
        k = 0.5
        starting_value_LC = 2.0
        user_specified_gain = 1.0

        A = pnl.TransferMechanism(function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic(gain=user_specified_gain), name='A')
        B = pnl.TransferMechanism(function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic(gain=user_specified_gain), name='B')
        C = pnl.Composition()
        LC = pnl.LCControlMechanism(
            modulated_mechanisms=[A, B],
            base_level_gain=G,
            scaling_factor_gain=k,
            objective_mechanism=pnl.ObjectiveMechanism(
                function=psyneulink.core.components.functions.nonstateful.transferfunctions.Linear,
                monitor=[B],
                name='LC ObjectiveMechanism'
            )
        )
        C.add_linear_processing_pathway([A,B])
        C.add_controller(LC)

        for output_port in LC.output_ports:
            output_port.parameters.value.set(output_port.value * starting_value_LC, C, override=True)

        LC.reset_stateful_function_when = pnl.Never()

        gain_created_by_LC_output_port_1 = []
        mod_gain_assigned_to_A = []
        base_gain_assigned_to_A = []
        mod_gain_assigned_to_B = []
        base_gain_assigned_to_B = []

        def report_trial(composition):
            from psyneulink import parse_context
            context = parse_context(composition)
            gain_created_by_LC_output_port_1.append(LC.output_ports[0].parameters.value.get(context))
            mod_gain_assigned_to_A.append([A.get_mod_gain(composition)])
            mod_gain_assigned_to_B.append([B.get_mod_gain(composition)])
            base_gain_assigned_to_A.append(A.function.gain.base)
            base_gain_assigned_to_B.append(B.function.gain.base)

        C._analyze_graph()
        benchmark(C.run, inputs={A: [[1.0], [1.0], [1.0], [1.0], [1.0]]},
              call_after_trial=functools.partial(report_trial, C))

        # (1) First value of gain in mechanisms A and B must be whatever we hardcoded for LC starting value
        assert mod_gain_assigned_to_A[0] == [starting_value_LC]

        # (2) _gain should always be set to user-specified value
        for i in range(5):
            assert base_gain_assigned_to_A[i] == user_specified_gain
            assert base_gain_assigned_to_B[i] == user_specified_gain

        # (3) LC output on trial n becomes gain of A and B on trial n + 1
        np.testing.assert_allclose(mod_gain_assigned_to_A[1:], gain_created_by_LC_output_port_1[0:-1])

        # (4) mechanisms A and B should always have the same gain values (b/c they are identical)
        np.testing.assert_allclose(mod_gain_assigned_to_A, mod_gain_assigned_to_B)

    @pytest.mark.mechanism
    @pytest.mark.control_mechanism
    @pytest.mark.benchmark(group="LCControlMechanism Basic")
    def test_lc_control_mech_basic(self, benchmark, mech_mode):

        LC = pnl.LCControlMechanism(
            base_level_gain=3.0,
            scaling_factor_gain=0.5,
            default_variable = 10.0
        )
        EX = pytest.helpers.get_mech_execution(LC, mech_mode)

        val = benchmark(EX, [10.0])
        expected = [[3.001397762387422]]
        # The difference in result shape is caused by shape mismatch in output port values.
        # The default shape is 1D, giving 2D overall result in compiled mode.
        # The true results are 2D per port, giving 3d overall result in Python mode.
        if mech_mode == 'Python':
            expected = [[ex] for ex in expected]
        np.testing.assert_allclose(val, expected)

    @pytest.mark.composition
    def test_lc_control_monitored_and_modulated_mechanisms_composition(self):
        """Test default configuration of LCControlMechanism with monitored and modulated mechanisms in a Composition
        Test that it implements an ObjectiveMechanism by default, that uses CombineMeans
        Test that ObjectiveMechanism can monitor Mechanisms with values of different lengths,
             and generate a scalar output.
        Test that it modulates all of the ProcessingMechanisms in the Composition but not the ObjectiveMechanism
        """

        T_1 = pnl.TransferMechanism(name='T_1', input_shapes=2)
        T_2 = pnl.TransferMechanism(name='T_2', input_shapes=3)

        C = pnl.Composition(pathways=[T_1, np.array([[1,2,3],[4,5,6]]), T_2])
        LC = pnl.LCControlMechanism(monitor_for_control=[T_1, T_2],
                                    modulated_mechanisms=C)
        C.add_node(LC)
        assert len(LC.control_signals)==1
        assert len(LC.control_signals[0].efferents)==2
        assert LC.path_afferents[0].sender.owner == LC.objective_mechanism
        assert isinstance(LC.objective_mechanism.function, pnl.CombineMeans)
        assert len(LC.objective_mechanism.input_ports[0].value) == 2
        assert len(LC.objective_mechanism.input_ports[1].value) == 3
        assert len(LC.objective_mechanism.output_ports[pnl.OUTCOME].value) == 1
        assert T_1.parameter_ports[pnl.SLOPE].mod_afferents[0] in LC.control_signals[0].efferents
        assert T_2.parameter_ports[pnl.SLOPE].mod_afferents[0] in LC.control_signals[0].efferents

        result = C.run(inputs={T_1:[1,2]})#, T_2:[3,4,5]
        assert LC.objective_mechanism.value == (np.mean(T_1.value) + np.mean(T_2.value))


@pytest.mark.composition
class TestControlMechanism:

    def test_control_modulation(self):
        Tx = pnl.TransferMechanism(name='Tx')
        Ty = pnl.TransferMechanism(name='Ty')
        Tz = pnl.TransferMechanism(name='Tz')
        C = pnl.ControlMechanism(
                # function=pnl.Linear,
                default_variable=[1],
                monitor_for_control=Ty,
                objective_mechanism=True,
                control_signals=pnl.ControlSignal(modulation=pnl.OVERRIDE,
                                                  modulates=(pnl.SLOPE, Tz)))
        comp=pnl.Composition(pathways=[[Tx, Tz],[Ty, C]])
        # comp.show_graph()

        assert Tz.parameter_ports[pnl.SLOPE].mod_afferents[0].sender.owner == C
        assert C.parameters.control_allocation.get() == [1]
        result = comp.run(inputs={Tx:[1,1], Ty:[4,4]})
        np.testing.assert_array_equal(comp.results, [[[4.], [4.]], [[4.], [4.]]])


    def test_identicalness_of_control_and_gating(self):
        """Tests same configuration as gating in tests/mechansims/test_gating_mechanism"""
        Input_Layer = pnl.TransferMechanism(name='Input Layer', function=pnl.Logistic, input_shapes=2)
        Hidden_Layer_1 = pnl.TransferMechanism(name='Hidden Layer_1', function=pnl.Logistic, input_shapes=5)
        Hidden_Layer_2 = pnl.TransferMechanism(name='Hidden Layer_2', function=pnl.Logistic, input_shapes=4)
        Output_Layer = pnl.TransferMechanism(name='Output Layer', function=pnl.Logistic, input_shapes=3)

        Control_Mechanism = pnl.ControlMechanism(
            input_shapes=[1], control=[Hidden_Layer_1.input_port,
                                       Hidden_Layer_2.input_port,
                                       Output_Layer.input_port])

        Input_Weights_matrix = (np.arange(2 * 5).reshape((2, 5)) + 1) / (2 * 5)
        Middle_Weights_matrix = (np.arange(5 * 4).reshape((5, 4)) + 1) / (5 * 4)
        Output_Weights_matrix = (np.arange(4 * 3).reshape((4, 3)) + 1) / (4 * 3)
        # This projection is specified in add_backpropagation_learning_pathway method below
        Input_Weights = pnl.MappingProjection(name='Input Weights',matrix=Input_Weights_matrix)
        # This projection is "discovered" by add_backpropagation_learning_pathway method below
        Middle_Weights = pnl.MappingProjection(name='Middle Weights',sender=Hidden_Layer_1,receiver=Hidden_Layer_2,
            matrix={
                pnl.VALUE: Middle_Weights_matrix,
                pnl.FUNCTION: pnl.AccumulatorIntegrator,
                pnl.FUNCTION_PARAMS: {
                    pnl.DEFAULT_VARIABLE: Middle_Weights_matrix,
                    pnl.INITIALIZER: Middle_Weights_matrix,
                    pnl.RATE: Middle_Weights_matrix
                },
            }
        )
        Output_Weights = pnl.MappingProjection(sender=Hidden_Layer_2,
                                           receiver=Output_Layer,
                                           matrix=Output_Weights_matrix)

        pathway = [Input_Layer, Input_Weights, Hidden_Layer_1, Hidden_Layer_2, Output_Layer]
        comp = pnl.Composition()
        backprop_pathway = comp.add_backpropagation_learning_pathway(
            pathway=pathway,
            loss_spec=None,
        )
        # c.add_linear_processing_pathway(pathway=z)
        comp.add_node(Control_Mechanism)

        np.testing.assert_allclose(Control_Mechanism.parameters.control_allocation.get(), [[0], [0], [0]])

        stim_list = {
            Input_Layer: [[-1, 30]],
            Control_Mechanism: [1.0],
            backprop_pathway.target: [[0, 0, 1]]}

        comp.learn(num_trials=3, inputs=stim_list)

        expected_results =[[[0.81493513, 0.85129046, 0.88154205]],
                           [[0.81331773, 0.85008207, 0.88157851]],
                           [[0.81168332, 0.84886047, 0.88161468]]]
        np.testing.assert_allclose(comp.results, expected_results)

        stim_list[Control_Mechanism]=[0.0]
        results = comp.learn(num_trials=1, inputs=stim_list)
        expected_results = [[0.5, 0.5, 0.5]]
        np.testing.assert_allclose(results, expected_results)

        stim_list[Control_Mechanism]=[2.0]
        results = comp.learn(num_trials=1, inputs=stim_list)
        expected_results = [[0.96941429, 0.9837254 , 0.99217549]]
        np.testing.assert_allclose(results, expected_results)

    def test_control_of_all_input_ports(self, comp_mode):
        mech = pnl.ProcessingMechanism(input_ports=['A','B','C'])
        control_mech = pnl.ControlMechanism(control=mech.input_ports)
        comp = pnl.Composition()
        comp.add_nodes([(mech, pnl.NodeRole.INPUT), (control_mech, pnl.NodeRole.INPUT)])
        results = comp.run(inputs={mech:[[2],[2],[2]], control_mech:[2]}, num_trials=2, execution_mode=comp_mode)

        np.testing.assert_allclose(control_mech.parameters.control_allocation.get(), [[1], [1], [1]])
        np.testing.assert_allclose(results, [[4], [4], [4]])


    def test_control_of_all_output_ports(self, comp_mode):
        mech = pnl.ProcessingMechanism(output_ports=[{pnl.VARIABLE: (pnl.OWNER_VALUE, 0)},
                                                      {pnl.VARIABLE: (pnl.OWNER_VALUE, 0)},
                                                      {pnl.VARIABLE: (pnl.OWNER_VALUE, 0)}],)
        control_mech = pnl.ControlMechanism(control=mech.output_ports)
        comp = pnl.Composition()
        comp.add_nodes([(mech, pnl.NodeRole.INPUT), (control_mech, pnl.NodeRole.INPUT)])
        results = comp.run(inputs={mech:[[2]], control_mech:[3]}, num_trials=2, execution_mode=comp_mode)

        np.testing.assert_allclose(control_mech.parameters.control_allocation.get(), [[1], [1], [1]])
        np.testing.assert_allclose(results, [[6], [6], [6]])

    def test_control_signal_default_allocation_specification(self):

        m1 = pnl.ProcessingMechanism()
        m2 = pnl.ProcessingMechanism()
        m3 = pnl.ProcessingMechanism()

        # default_allocation *not* specified in constructor of pnl.ControlMechanism,
        #     so should be set to defaultControlAllocation (=[1])
        c1 = pnl.ControlMechanism(name='C1',
                                  default_variable=[10],
                                  control_signals=[pnl.ControlSignal(modulates=(pnl.SLOPE, m1)), # test for assignment to defaultControlAllocation
                                                   pnl.ControlSignal(default_allocation=2,       # test for scalar assignment
                                                                     modulates=(pnl.SLOPE, m2)),
                                                   pnl.ControlSignal(default_allocation=[3],     # test for array assignment
                                                                     modulates=(pnl.SLOPE, m3))])
        comp = pnl.Composition()
        comp.add_nodes([m1,m2,m3])
        comp.add_controller(c1)
        # Default controL_allocation should = defaultControlAllocation (since default_allocation arg was not specified)
        np.testing.assert_allclose(c1.defaults.control_allocation, [1])
        # Initial control_allocation should reflect the default input ([10])
        np.testing.assert_allclose(c1.parameters.control_allocation.get(), [[10], [10], [10]])
        # Initial values of the ControlSignals should reflect *their* default assignments (since not yet executed)
        # and parameters they modulate (SLOPE) should reflect *their* initial values (1)
        assert c1.control_signals[0].value == [1] # defaultControlAllocation ([1]) should be assigned,
                                                   # as no default_allocation from pnl.ControlMechanism
        assert m1.parameter_ports[pnl.SLOPE].value == [1]
        assert c1.control_signals[1].value == [2]      # default_allocation from pnl.ControlSignal (converted scalar)
        assert m2.parameter_ports[pnl.SLOPE].value == [1]
        assert c1.control_signals[2].value == [3]      # default_allocation from pnl.ControlSignal
        assert m3.parameter_ports[pnl.SLOPE].value == [1]
        result = comp.run(inputs={m1:[2],m2:[3],m3:[4]})
        # Result should reflect:
        # 1) use of initial ControlSignal values to set parmeters
        # 2) updating of the ControlSignal values ([10]'s) from ControlMechanism's default_variable
        #     (since it does not have any other source of input control_allocation)
        np.testing.assert_allclose(result, [[2.], [6.], [12.]])
        assert c1.control_signals[0].value == [10]
        assert m1.parameter_ports[pnl.SLOPE].value == [1]
        assert c1.control_signals[1].value == [10]
        assert m2.parameter_ports[pnl.SLOPE].value == [2]
        assert c1.control_signals[2].value == [10]
        assert m3.parameter_ports[pnl.SLOPE].value == [3]
        # Results should now reflect use of updated ControlSignal values to set parmeters
        result = comp.run(inputs={m1:[2],m2:[3],m3:[4]})
        np.testing.assert_allclose(result, [[20.], [30.], [40.]])
        assert c1.control_signals[0].value == [10]
        assert m1.parameter_ports[pnl.SLOPE].value == [10]
        assert c1.control_signals[1].value == [10]
        assert m2.parameter_ports[pnl.SLOPE].value == [10]
        assert c1.control_signals[2].value == [10]
        assert m3.parameter_ports[pnl.SLOPE].value == [10]

        # default_allocation *is* specified in constructor of pnl.ControlMechanism,
        #     so should be used unless specified in pnl.ControlSignal constructor
        c2 = pnl.ControlMechanism(
                name='C3',
                default_variable=[10],
                default_allocation=[4],
                # Test synonyms allowed for **control**: generic **modulates**, and even more generic **projections**):
                control_signals=[pnl.ControlSignal(control=(pnl.SLOPE, m1)),  # tests for assignment to default_allocation
                                 pnl.ControlSignal(default_allocation=5,  # tests for override of default_allocation
                                                   modulates=(pnl.SLOPE, m2)),
                                 pnl.ControlSignal(default_allocation=[6],  # as above same but with array
                                                   projections=(pnl.SLOPE, m3))])
        comp = pnl.Composition()
        comp.add_nodes([m1,m2,m3])
        comp.add_controller(c2)
        np.testing.assert_allclose(c2.parameters.control_allocation.get(), [[10], [10], [10]])
        assert c2.control_signals[0].value == [4]        # default_allocation from pnl.ControlMechanism assigned
        assert m1.parameter_ports[pnl.SLOPE].value == [10]  # has not yet received pnl.ControlSignal value
        assert c2.control_signals[1].value == [5]        # default_allocation from pnl.ControlSignal assigned (converted scalar)
        assert m2.parameter_ports[pnl.SLOPE].value == [10]
        assert c2.control_signals[2].value == [6]        # default_allocation from pnl.ControlSignal assigned
        assert m3.parameter_ports[pnl.SLOPE].value == [10]
        result = comp.run(inputs={m1:[2],m2:[3],m3:[4]})
        np.testing.assert_allclose(result, [[8.], [15.], [24.]])
        assert c2.control_signals[0].value == [10]
        assert m1.parameter_ports[pnl.SLOPE].value == [4]
        assert c2.control_signals[1].value == [10]
        assert m2.parameter_ports[pnl.SLOPE].value == [5]
        assert c2.control_signals[2].value == [10]
        assert m3.parameter_ports[pnl.SLOPE].value == [6]
        result = comp.run(inputs={m1:[2],m2:[3],m3:[4]})
        np.testing.assert_allclose(result, [[20.], [30.], [40.]])
        assert c2.control_signals[0].value == [10]
        assert m1.parameter_ports[pnl.SLOPE].value == [10]
        assert c2.control_signals[1].value == [10]
        assert m2.parameter_ports[pnl.SLOPE].value == [10]
        assert c2.control_signals[2].value == [10]
        assert m3.parameter_ports[pnl.SLOPE].value == [10]



import argparse

import numpy as np

import psyneulink as pnl

parser = argparse.ArgumentParser()
parser.add_argument('--no-plot', action='store_false', help='Disable plotting', dest='enable_plot')
parser.add_argument('--threshold', type=float, help='Termination threshold for response output (default: %(default)f)', default=0.55)
parser.add_argument('--word-runs', type=int, help='Number of runs after word is presented (default: %(default)d)', default=5)
parser.add_argument('--color-runs', type=int, help='Number of runs after color is presented (default: %(default)d)', default=4)
parser.add_argument('--settle-trials', type=int, help='Number of trials for composition to initialize and settle (default: %(default)d)', default=50)
parser.add_argument('--pre-stimulus-trials', type=int, help='Number of trials before stimulus is added', default=100)
args = parser.parse_args()

# This implements the horse race Figure shown in Cohen & Huston (1994).
# Note that noise is turned off and each stimulus is only showed once for each stimulus onset asynchrony.

# Define Variables ----------------------------------------------------------------------------------------------------
rate = 0.1          # The integration rate was changed from 0.01 to 0.1
inhibition = -2.0   # Mutual inhibition across each layer
bias = 4.0          # bias for hidden layer units
threshold = args.threshold    # Threshold until a response is made, changed from 0.6 to 0.55
settle_trials = args.settle_trials  # Time for system to initialize and settle
prior120 = args.pre_stimulus_trials      # Cycles needed to be added for stimulus to start

# Different time steps at which the System should end run and start new terminate_processing run
# This is needed for conditions in which the irrelevant condition is like a neutral trial and could already lead to
# a correct response. This is basically the reason why with long positive stimulus onset asynchrony the three
# condistions (congruent, incongruent, neutral) lead to the same reaction time.
terminate2 = 180
terminate3 = 200
terminate4 = 220
terminate5 = 240

# Create mechanisms ---------------------------------------------------------------------------------------------------
#   Linear input units, colors: ('red', 'green'), words: ('RED','GREEN')
colors_input_layer = pnl.TransferMechanism(
    input_shapes=3,
                                           function=pnl.Linear,
                                           name='COLORS_INPUT')

words_input_layer = pnl.TransferMechanism(
    input_shapes=3,
                                          function=pnl.Linear,
                                          name='WORDS_INPUT')

task_input_layer = pnl.TransferMechanism(
    input_shapes=2,
                                         function=pnl.Linear,
                                         name='TASK_INPUT')

#   Task layer, tasks: ('name the color', 'read the word')
task_layer = pnl.RecurrentTransferMechanism(
    input_shapes=2,
                                            function=pnl.Logistic(),
                                            hetero=-2,
                                            integrator_mode=True,
                                            integration_rate=0.1,
                                            name='TASK')

#   Hidden layer units, colors: ('red','green') words: ('RED','GREEN')
colors_hidden_layer = pnl.RecurrentTransferMechanism(
    input_shapes=3,
                                                     function=pnl
                                                     .Logistic(x_0=4.0),
                                                     integrator_mode=True,
                                                     hetero=-2.0,
                                                     # noise=pnl.NormalDist(mean=0.0, standard_deviation=.0),
                                                     integration_rate=0.1,  # cohen-huston text says 0.01
                                                     name='COLORS HIDDEN')

words_hidden_layer = pnl.RecurrentTransferMechanism(
    input_shapes=3,
                                                    function=pnl.Logistic(x_0=4.0),
                                                    hetero=-2,
                                                    integrator_mode=True,
                                                    # noise=pnl.NormalDist(mean=0.0, standard_deviation=.05),
                                                    integration_rate=0.1,
                                                    name='WORDS HIDDEN')
#   Response layer, responses: ('red', 'green'): RecurrentTransferMechanism for self inhibition matrix
response_layer = pnl.RecurrentTransferMechanism(
    input_shapes=2,
                                                function=pnl.Logistic(),
                                                hetero=-2.0,
                                                integrator_mode=True,
                                                integration_rate=0.1,
                                                name='RESPONSE')

# Log mechanisms ------------------------------------------------------------------------------------------------------
#task_layer.set_log_conditions('gain')
task_layer.set_log_conditions('value')
colors_hidden_layer.set_log_conditions('value')
words_hidden_layer.set_log_conditions('value')
response_layer.set_log_conditions('value')
# Connect mechanisms --------------------------------------------------------------------------------------------------
# (note that response layer projections are set to all zero first for initialization

color_input_weights = pnl.MappingProjection(matrix=np.array([[1.0, 0.0, 0.0],
                                                             [0.0, 1.0, 0.0],
                                                             [0.0, 0.0, 0.0]]))

word_input_weights = pnl.MappingProjection(matrix=np.array([[1.0, 0.0, 0.0],
                                                            [0.0, 1.0, 0.0],
                                                            [0.0, 0.0, 0.0]]))

task_input_weights = pnl.MappingProjection(matrix=np.array([[1.0, 0.0],
                                                            [0.0, 1.0]]))

color_task_weights  = pnl.MappingProjection(matrix=np.array([[4.0, 0.0],
                                                             [4.0, 0.0],
                                                             [4.0, 0.0]]))

task_color_weights  = pnl.MappingProjection(matrix=np.array([[4.0, 4.0, 4.0],
                                                             [0.0, 0.0, 0.0]]))

word_task_weights = pnl.MappingProjection(matrix=np.array([[0.0, 4.0],
                                                           [0.0, 4.0],
                                                           [0.0, 4.0]]))

task_word_weights = pnl.MappingProjection(matrix=np.array([[0.0, 0.0, 0.0],
                                                           [4.0, 4.0, 4.0]]))

response_color_weights = pnl.MappingProjection(matrix=np.array([[0.0, 0.0, 0.0],
                                                                [0.0, 0.0, 0.0]]))

response_word_weights  = pnl.MappingProjection(matrix=np.array([[0.0, 0.0, 0.0],
                                                                [0.0, 0.0, 0.0]]))

color_response_weights = pnl.MappingProjection(matrix=np.array([[1.5, 0.0],
                                                                [0.0, 1.5],
                                                                [0.0, 0.0]]))
word_response_weights  = pnl.MappingProjection(matrix=np.array([[2.5, 0.0],
                                                                [0.0, 2.5],
                                                                [0.0, 0.0]]))

Bidirectional_Stroop = pnl.Composition(name='FEEDFORWARD_STROOP_SYSTEM')

# Create pathways -----------------------------------------------------------------------------------------------------
Bidirectional_Stroop.add_linear_processing_pathway(
    pathway=[
        colors_input_layer,
        color_input_weights,
        colors_hidden_layer,
        color_response_weights,
        response_layer,
        response_color_weights,
        colors_hidden_layer
    ],
    name='COLORS_RESPONSE_PROCESS'
)

Bidirectional_Stroop.add_linear_processing_pathway(
    pathway=[
        words_input_layer,
        word_input_weights,
        words_hidden_layer,
        word_response_weights,
        response_layer,
        response_word_weights,
        words_hidden_layer
    ],
    name='WORDS_RESPONSE_PROCESS'
)

Bidirectional_Stroop.add_linear_processing_pathway(
    pathway=[
        task_input_layer,
        task_input_weights,
        task_layer,
        task_color_weights,
        colors_hidden_layer,
        color_task_weights,
        task_layer
    ]
)

Bidirectional_Stroop.add_linear_processing_pathway(
    pathway=[
        task_input_layer,
        task_layer,
        task_word_weights,
        words_hidden_layer,
        word_task_weights,
        task_layer
    ]
)

# LOGGING:
colors_hidden_layer.set_log_conditions('value')
words_hidden_layer.set_log_conditions('value')

# Bidirectional_Stroop.show_graph(show_dimensions=pnl.ALL)#,show_mechanism_structure=pnl.VALUES) # Uncomment to show graph of the system


def terminate_num_executions(length, context):
    return response_layer.log.nparray_dictionary()[context.execution_id]['value'].shape[0] == length


# Create different terminate trial conditions --------------------------------------------------------------------------
terminate_trial = {
    pnl.TimeScale.TRIAL: pnl.Or(
        pnl.Threshold(response_layer, 'value', threshold, '>=', (0, 0)),
        pnl.Threshold(response_layer, 'value', threshold, '>=', (0, 1)),
    )
}
terminate_trial2 = {
    pnl.TimeScale.TRIAL: pnl.Or(
        pnl.Threshold(response_layer, 'value', threshold, '>=', (0, 0)),
        pnl.Threshold(response_layer, 'value', threshold, '>=', (0, 1)),
        pnl.While(terminate_num_executions, terminate2),
    )
}
terminate_trial3 = {
    pnl.TimeScale.TRIAL: pnl.Or(
        pnl.Threshold(response_layer, 'value', threshold, '>=', (0, 0)),
        pnl.Threshold(response_layer, 'value', threshold, '>=', (0, 1)),
        pnl.While(terminate_num_executions, terminate3),
    )
}
terminate_trial4 = {
    pnl.TimeScale.TRIAL: pnl.Or(
        pnl.Threshold(response_layer, 'value', threshold, '>=', (0, 0)),
        pnl.Threshold(response_layer, 'value', threshold, '>=', (0, 1)),
        pnl.While(terminate_num_executions, terminate4),
    )
}
terminate_trial5 = {
    pnl.TimeScale.TRIAL: pnl.Or(
        pnl.Threshold(response_layer, 'value', threshold, '>=', (0, 0)),
        pnl.Threshold(response_layer, 'value', threshold, '>=', (0, 1)),
        pnl.While(terminate_num_executions, terminate5),
    )
}

terminate_list = [terminate_trial2,
                  terminate_trial3,
                  terminate_trial4,
                  terminate_trial5]

# Create test trials function -----------------------------------------------------------------------------------------
# a BLUE word input is [1,0] to words_input_layer and GREEN word is [0,1]
# a blue color input is [1,0] to colors_input_layer and green color is [0,1]
# a color-naming trial is [1,0] to task_layer and a word-reading trial is [0,1]
def trial_dict(red_color, green_color, neutral_color, red_word, green_word, neutral_word, CN, WR):

    trialdict = {
    colors_input_layer: [red_color, green_color, neutral_color],
    words_input_layer: [red_word, green_word, neutral_word],
    task_input_layer: [CN, WR]
    }
    return trialdict


# Define initialization trials separately
# WR_initialize_input = trial_dict(0, 0, 0, 0, 0, 0, 0, 1)
CN_initialize_input = trial_dict(0, 0, 0, 0, 0, 0, 1, 0)

CN_incongruent_trial_input = trial_dict(1, 0, 0, 0, 1, 0, 1, 0) #red_color, green color, red_word, green word, CN, WR
CN_congruent_trial_input = trial_dict(1, 0, 0, 1, 0, 0, 1, 0) #red_color, green color, red_word, green word, CN, WR
CN_control_trial_input = trial_dict(1, 0, 0, 0, 0, 0, 1, 0) #red_color, green color, red_word, green word, CN, WR
CN_control_word_trial_input = trial_dict(0, 0, 0, 1, 0, 0, 1, 0) #red_color, green color, red_word, green word, CN, WR

CN_congruent_word_first_input = trial_dict(0, 0, 0, 1, 0, 0, 1, 0) #red_color, green color, red_word, green word, CN, WR
CN_incongruent_word_first_input = trial_dict(0, 0, 0, 0, 1, 0, 1, 0) #red_color, green color, red_word, green word, CN, WR

# WR_congruent_trial_input = trial_dict(1, 0, 0, 1, 0, 0, 0, 1) #red_color, green color, red_word, green word, CN, WR
# WR_incongruent_trial_input = trial_dict(1, 0, 0, 0, 1, 0, 0, 1) #red_color, green color, red_word, green word, CN, WR
# WR_control_trial_input = trial_dict(1, 0, 0, 0, 0, 0, 0, 1) #red_color, green color, red_word, green word, CN, WR

conditions = 3
runs = args.word_runs
runs2 = args.color_runs
response_all = []
response_all2 = []

Stimulus = [[CN_initialize_input, CN_congruent_word_first_input, CN_congruent_trial_input, CN_control_trial_input],
            [CN_initialize_input, CN_incongruent_word_first_input, CN_incongruent_trial_input, CN_control_trial_input],
            [CN_initialize_input, CN_control_word_trial_input, CN_control_trial_input, CN_control_trial_input]]


post_settlement_multiplier = int(prior120 / 5)

# First "for loop" over conditions
# Second "for loop" over runs
for cond in range(conditions):
# ---------------------------------------------------------------------------------------------------------------------
    # Run congruent trial with word presented 1200 trials prior ------------------------------------------------------------
    for run in range(runs):  # use 1 as starting point to insure Composition is called with non-zero num_trials)
        response_color_weights = pnl.MappingProjection(matrix=np.array([[0.0, 0.0, 0.0],
                                                                        [0.0, 0.0, 0.0]]))

        response_word_weights = pnl.MappingProjection(matrix=np.array([[0.0, 0.0, 0.0],
                                                                       [0.0, 0.0, 0.0]]))
        Bidirectional_Stroop.run(inputs=Stimulus[cond][0], num_trials=settle_trials)    # run system to settle for 200 trials with congruent stimuli input
        Bidirectional_Stroop.run(inputs=Stimulus[cond][0], num_trials=post_settlement_multiplier * (run))  # run system to settle for 200 trials with congruent stimuli input

        response_color_weights = pnl.MappingProjection(matrix=np.array([[1.5, 0.0, 0.0],
                                                                        [0.0, 1.5, 0.0]]))

        response_word_weights  = pnl.MappingProjection(matrix=np.array([[2.5, 0.0, 0.0],
                                                                        [0.0, 2.5, 0.0]]))

        Bidirectional_Stroop.run(inputs=Stimulus[cond][1], num_trials=prior120 - (run * post_settlement_multiplier))# termination_processing=terminate_trial) # run system with congruent stimulus input until
        Bidirectional_Stroop.run(inputs=Stimulus[cond][2], termination_processing=terminate_trial) # run system with congruent stimulus input until
                                                                    # threshold in of of the response layer units is reached

    # Store values from run -----------------------------------------------------------------------------------------------
        r = response_layer.log.nparray_dictionary('value')       # Log response output from special logistic function
        rr = r[Bidirectional_Stroop.name]['value']
        n_r = rr.shape[0]
        rrr = rr.reshape(n_r,2)

        response_all.append(rrr.shape[0])

        # Clear log & reset ----------------------------------------------------------------------------------------
        response_layer.log.clear_entries()
        colors_hidden_layer.log.clear_entries()
        words_hidden_layer.log.clear_entries()
        task_layer.log.clear_entries()

        colors_hidden_layer.reset([[0, 0, 0]], context=Bidirectional_Stroop)
        words_hidden_layer.reset([[0, 0, 0]], context=Bidirectional_Stroop)
        response_layer.reset([[0, 0]], context=Bidirectional_Stroop)
        task_layer.reset([[0, 0]], context=Bidirectional_Stroop)

    print('response_all: ', response_all)

    # Run trials after congruent color was presented ----------------------------------------------------------------------
    for run2 in range(runs2):
        response_color_weights = pnl.MappingProjection(matrix=np.array([[0.0, 0.0, 0.0],
                                                                        [0.0, 0.0, 0.0]]))
        response_word_weights = pnl.MappingProjection(matrix=np.array([[0.0, 0.0, 0.0],
                                                                       [0.0, 0.0, 0.0]]))
        Bidirectional_Stroop.run(inputs=Stimulus[cond][0], num_trials = settle_trials)  # run system to settle for 200 trials with congruent stimuli input
        Bidirectional_Stroop.run(inputs=Stimulus[cond][0], num_trials = prior120)  # run system to settle for 200 trials with congruent stimuli input
        response_color_weights = pnl.MappingProjection(matrix=np.array([[1.5, 0.0, 0.0],
                                                                        [0.0, 1.5, 0.0]]))
        response_word_weights = pnl.MappingProjection(matrix=np.array([[2.5, 0.0, 0.0],
                                                                       [0.0, 2.5, 0.0]]))
        Bidirectional_Stroop.run(inputs=Stimulus[cond][3], termination_processing=terminate_list[run2])#terminate_list[run2])  # termination_processing=terminate_trial) # run system with congruent stimulus input until

        Bidirectional_Stroop.run(inputs=Stimulus[cond][2], termination_processing=terminate_trial)  # run system with congruent stimulus input until
        # threshold in of of the response layer units is reached
        # Store values from run -----------------------------------------------------------------------------------------------
        r = response_layer.log.nparray_dictionary('value')       # Log response output from special logistic function
        rr = r[Bidirectional_Stroop.name]['value']
        n_r = rr.shape[0]
        rrr = rr.reshape(n_r,2)
        response_all.append(rrr.shape[0])
        # Clear log & reset ------------------------------------------------------------------------------------
        response_layer.log.clear_entries()
        colors_hidden_layer.log.clear_entries()
        words_hidden_layer.log.clear_entries()
        task_layer.log.clear_entries()
        colors_hidden_layer.reset([[0, 0, 0]], context=Bidirectional_Stroop)
        words_hidden_layer.reset([[0, 0, 0]], context=Bidirectional_Stroop)
        response_layer.reset([[0, 0]], context=Bidirectional_Stroop)
        task_layer.reset([[0, 0]], context=Bidirectional_Stroop)

# Plotting ------------------------------------------------------------------------------------------------------------
if args.enable_plot:
    import matplotlib.pyplot as plt

    # compute regression for model
    reg = np.dot(response_all, 2) + 123
    plt.figure()
    # plt.plot(response_all[0:9])
    # plt.plot(response_all[9:18])
    # plt.plot(response_all[18:27])

    response_len = runs + runs2

    stimulus_onset_asynchrony = np.linspace(-400, 400, response_len)
    plt.plot(stimulus_onset_asynchrony, reg[0:response_len], '-^')
    plt.plot(stimulus_onset_asynchrony, reg[response_len:2 * response_len], '-s')
    plt.plot(stimulus_onset_asynchrony, reg[2 * response_len:3 * response_len], '-o')
    plt.title('stimulus onset asynchrony - horse race model ')
    plt.legend(['congruent', 'incongruent', 'neutral'])
    plt.ylabel('reaction time in ms')
    plt.show(block=not pnl._called_from_pytest)



import numpy as np
import psyneulink as pnl

# Input Ports exposed to user, with default values
hid1_layer_weight = np.zeros((90, 90), dtype="float32")
input1 = np.zeros((90,), dtype="float32")
embed_bias_weight = np.zeros((10, 90), dtype="float32")
hid2_layer_weight = np.zeros((90, 80), dtype="float32")
out_layer_weight = np.zeros((80, 2), dtype="float32")

FFWMGraph = pnl.Composition(name="FFWMGraph")

Add_10 = pnl.ProcessingMechanism(
    name="Add_10",
    function=pnl.Identity(default_variable=np.zeros((2, 90), dtype="float32")),
    default_variable=np.zeros((2, 90), dtype="float32"),
)
Dropout_13_14 = pnl.ProcessingMechanism(
    name="Dropout_13_14",
    function=pnl.Dropout(
        p=0.05000000074505806, default_variable=np.zeros((90,), dtype="float32")
    ),
    default_variable=np.zeros((90,), dtype="float32"),
)
Gather_9 = pnl.ProcessingMechanism(
    name="Gather_9",
    function=pnl.Identity(default_variable=np.zeros((10, 90), dtype="float32")),
    default_variable=np.zeros((10, 90), dtype="float32"),
)
MatMul_19_passthrough_terminal = pnl.ProcessingMechanism(
    name="MatMul_19_passthrough_terminal",
    function=pnl.Identity(default_variable=[0.0, 0.0]),
    default_variable=[0.0, 0.0],
)
MatMul_6_passthrough_origin = pnl.ProcessingMechanism(
    name="MatMul_6_passthrough_origin",
    function=pnl.Identity(default_variable=np.zeros((90,), dtype="float32")),
    default_variable=np.zeros((90,), dtype="float32"),
)
Relu_17 = pnl.ProcessingMechanism(
    name="Relu_17",
    function=pnl.ReLU(default_variable=np.zeros((80,), dtype="float32")),
    default_variable=np.zeros((80,), dtype="float32"),
)
Relu_7 = pnl.ProcessingMechanism(
    name="Relu_7",
    function=pnl.ReLU(default_variable=np.zeros((90,), dtype="float32")),
    default_variable=np.zeros((90,), dtype="float32"),
)

FFWMGraph.add_node(Add_10)
FFWMGraph.add_node(Dropout_13_14)
FFWMGraph.add_node(Gather_9)
FFWMGraph.add_node(MatMul_19_passthrough_terminal)
FFWMGraph.add_node(MatMul_6_passthrough_origin)
FFWMGraph.add_node(Relu_17)
FFWMGraph.add_node(Relu_7)

FFWMGraph.add_projection(
    projection=pnl.MappingProjection(name="Relu_7_Add_10"),
    sender=Relu_7,
    receiver=Add_10,
)
FFWMGraph.add_projection(
    projection=pnl.MappingProjection(name="Gather_9_Add_10"),
    sender=Gather_9,
    receiver=Add_10,
)
FFWMGraph.add_projection(
    projection=pnl.MappingProjection(name="Add_10_Dropout_13_14"),
    sender=Add_10,
    receiver=Dropout_13_14,
)
FFWMGraph.add_projection(
    projection=pnl.MappingProjection(
        name="MatMul_6_as_edge",
        function=pnl.MatrixTransform(
            default_variable=np.zeros((90,), dtype="float32"), matrix=hid1_layer_weight
        ),
    ),
    sender=MatMul_6_passthrough_origin,
    receiver=Relu_7,
)
FFWMGraph.add_projection(
    projection=pnl.MappingProjection(
        name="MatMul_16_as_edge",
        function=pnl.MatrixTransform(
            default_variable=np.zeros((90,), dtype="float32"), matrix=hid2_layer_weight
        ),
    ),
    sender=Dropout_13_14,
    receiver=Relu_17,
)
FFWMGraph.add_projection(
    projection=pnl.MappingProjection(
        name="MatMul_19_as_edge",
        function=pnl.MatrixTransform(
            default_variable=np.zeros((80,), dtype="float32"), matrix=out_layer_weight
        ),
    ),
    sender=Relu_17,
    receiver=MatMul_19_passthrough_terminal,
)


from . import autoassociativelearningmechanism
from . import kohonenlearningmechanism
from . import EMstoragemechanism

from .autoassociativelearningmechanism import *
from .kohonenlearningmechanism import *
from .EMstoragemechanism import *

__all__ = list(autoassociativelearningmechanism.__all__)
__all__.extend(kohonenlearningmechanism.__all__)
__all__.extend(EMstoragemechanism.__all__)
